from __future__ import annotations

from dagster import (
    AssetKey,
    AssetSelection,
    DagsterInstance,
    DefaultSensorStatus,
    DynamicPartitionsDefinition,
    Output,
    RunRequest,
    asset,
    multi_asset_sensor,
)
from dagster._core.errors import DagsterHomeNotSetError


class CloudAssetSensor:
    """
    Class which scaffolds a cloud sensor. This class defines a publishing
    asset, which uses a custom IO manager to publish data either to Azure or
    local storage.

    The publishing asset in turn monitors a list of country-specific assets,
    which are responsible for generating the data that are to be published.

    Arguments
    ---------
    `io_manager_key`: str
        The key of the IO manager used for publishing. See the 'resources' and
        'resources_by_env' dicts in python/poppusher/__init__.py.

    `prefix`: str
        A string used to disambiguate the different assets / sensors that are
        being generated here, as Dagster does not allow duplicated names.

    `interval`: int
        The minimum interval at which the sensor should run, in seconds.
    """

    def __init__(
        self,
        io_manager_key: str,
        prefix: str,
        interval: int,
    ):
        self.io_manager_key = io_manager_key
        self.publishing_asset_name = f"publish_{prefix}"
        self.sensor_name = f"sensor_{prefix}"
        self.interval = interval
        self.monitored_asset_keys: list[AssetKey] = []

        self.partition_definition_name = f"{prefix}_partitions"
        self.partition_definition = DynamicPartitionsDefinition(
            name=self.partition_definition_name
        )
        # Upon initialisation, remove all partitions so that the web UI doesn't
        # show any stray partitions left over from old Dagster launches (this
        # can happen when e.g. switching between different git branches)
        try:
            with DagsterInstance.get() as instance:
                for partition_key in instance.get_dynamic_partitions(
                    self.partition_definition_name
                ):
                    instance.delete_dynamic_partition(
                        self.partition_definition_name, partition_key
                    )
        except DagsterHomeNotSetError:
            pass

    def create_publishing_asset(self):
        @asset(
            name=self.publishing_asset_name,
            partitions_def=self.partition_definition,
            io_manager_key=self.io_manager_key,
        )
        def publish(context):
            from poppusher import defs as poppusher_defs

            # load_asset_value expects a list of strings
            output = poppusher_defs.load_asset_value(context.partition_key.split("/"))
            return Output(output)

        return publish

    def create_sensor(self):
        @multi_asset_sensor(
            # Because the list of monitored assets is dynamically generated
            # using self.monitored_asset_keys, we don't pass it here.
            monitored_assets=[],
            request_assets=AssetSelection.keys(self.publishing_asset_name),
            name=self.sensor_name,
            minimum_interval_seconds=self.interval,
            default_status=DefaultSensorStatus.RUNNING,
        )
        def inner_sensor(context):
            # Get info about the latest materialisations of all the assets
            # we're interested in
            asset_events = context.latest_materialization_records_by_key(
                self.monitored_asset_keys
            )

            for monitored_asset_key, execution_value in asset_events.items():
                monitored_asset_name = "/".join(monitored_asset_key.path)

                # Add the monitored asset to the list of dynamic partitions of
                # the publishing asset, if it's not already there.
                if monitored_asset_name not in context.instance.get_dynamic_partitions(
                    self.partition_definition_name
                ):
                    context.instance.add_dynamic_partitions(
                        partitions_def_name=self.partition_definition_name,
                        partition_keys=[monitored_asset_name],
                    )

                if execution_value is not None:
                    # Trigger a run request for the publishing asset, if it has
                    # been materialised.
                    # execution_value.run_id contains the run ID of the last time
                    # the monitored asset was materialised. By setting the run_key
                    # to this, we ensure we don't run the publishing sensor
                    # multiple times for the same asset materialisation.
                    yield RunRequest(
                        run_key=f"{monitored_asset_name}_{execution_value.run_id}",
                        partition_key="/".join(monitored_asset_key.path),
                    )
                    context.advance_cursor({monitored_asset_key: execution_value})

        return inner_sensor



from __future__ import annotations

import poppusher._core as m


def test_add():
    assert m.add(2, 3) == 5


def test_subtract():
    assert m.subtract(7, 5) == 2


def test_multiply():
    assert m.multiply(7, 5) == 35



from __future__ import annotations

from pathlib import Path

import pandas as pd
from dagster import build_asset_context
from icecream import ic

from poppusher.assets.gb_eaw import england_wales_census as ew_census
from poppusher.metadata import MetricMetadata


def test_retrieve_table_description():
    example_url = "https://www.nomisweb.co.uk/datasets/c2021ts009"
    expected_desction_snipet = "estimates that classify usual residents in England and Wales by sex and single year of age"
    actual_description = ew_census._retrieve_table_description(example_url)

    assert (
        expected_desction_snipet in actual_description
    ), "The description of the table did not included the expected text"


def test_uk__derived_metrics():
    mmd = MetricMetadata(
        human_readable_name="test_human_readable_name",
        source_download_url="test_source_download_url",
        source_archive_file_path="test_source_archive_file_path",
        source_documentation_url="test_source_documentation_url",
        source_data_release_id="geography code",
        # TODO - this is a placeholder
        parent_metric_id="unknown_at_this_stage",
        potential_denominator_ids=None,
        parquet_margin_of_error_file=None,
        parquet_margin_of_error_column=None,
        parquet_column_name="unknown_parquet_column_name",
        # TODO - this is a placeholder
        metric_parquet_path="unknown",
        hxl_tag="hxl_tag_unknown",
        description="description",
        source_metric_id="source_table.hxltag",
    )

    # Get a context for testing
    context = build_asset_context(partition_key="ltla/TS009")

    # TODO, replace this with a proper fixture
    demo_census_source_table_path = (
        Path(__file__).parent / "demo_data" / "gbr_ew_census2021-ts009-ltla.csv"
    )
    source_df = pd.read_csv(demo_census_source_table_path)

    # Limit the test data to just Hartlepool and Middlesbrough
    # - Two rows in the test data, to distinguish between other cases where derived
    # metrics are calculated by pivoting over multiple rows
    # - Still dealing with the source table here, so the relevant columns is "geography code"
    source_df = source_df[source_df["geography code"].isin(["E06000001", "E06000002"])]

    ewc = ew_census.EnglandAndWales()
    actual_derived_metrics = ewc._derived_metrics(context, source_df, mmd)

    # ic(actual_derived_metrics)
    ic(actual_derived_metrics.metrics.head(1).T)

    # Test the results on Hartlepool
    # - This is the derived metrics, hence the relevant column is "GEO_ID"
    hartlepool = actual_derived_metrics.metrics[
        actual_derived_metrics.metrics["GEO_ID"] == "E06000001"
    ]

    # Manually calculated expected values
    assert hartlepool["children_5_17"].to_numpy()[0] == 14847
    assert hartlepool["infants_0_4"].to_numpy()[0] == 4981

    assert hartlepool["children_0_17"].to_numpy()[0] == 19828
    assert hartlepool["adults_f"].to_numpy()[0] == 37971
    assert hartlepool["adults_m"].to_numpy()[0] == 34548
    assert hartlepool["adults"].to_numpy()[0] == 72519
    assert hartlepool["individuals"].to_numpy()[0] == 92347



from __future__ import annotations

import tempfile
import uuid
from pathlib import Path

import geopandas as gpd
from dagster import (
    AssetKey,
    AssetSelection,
    Config,
    DefaultSensorStatus,
    DynamicPartitionsDefinition,
    Noneable,
    Output,
    RunConfig,
    RunRequest,
    SkipReason,
    asset,
    define_asset_job,
    load_assets_from_current_module,
    multi_asset_sensor,
)
from icecream import ic
from slugify import slugify

cloud_assets = load_assets_from_current_module(group_name="cloud_assets")


publishing_partition = DynamicPartitionsDefinition(name="publishing_partition")


cloud_assets_job = define_asset_job(
    name="cloud_assets",
    selection=AssetSelection.groups("cloud_assets"),
    partitions_def=publishing_partition,
)


class CountryAssetConfig(Config):
    asset_to_load: str
    partition_to_load: str | None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


# TODO. This is a temporary structure to list the end points of each country pipeline
# A more robust solution should be implemented as part of
# https://github.com/Urban-Analytics-Technology-Platform/poppusher/issues/38
# and https://github.com/Urban-Analytics-Technology-Platform/poppusher/issues/39
assets_to_monitor = (
    # UK Geography
    (
        AssetSelection.groups("uk")
        & AssetSelection.keys("boundary_line").downstream(include_self=False)
    )
    # USA Geography
    | (
        AssetSelection.groups("us")
        & AssetSelection.keys("geometry_ids").downstream(include_self=False)
    )
    # Belgium Geography + tables
    | (
        AssetSelection.groups("be")
        # Note: include_self=True as the geodataframe is currently not output for
        # derived tables but only for source table:
        # be-source-table-https-statbel-fgov-be-node-4726
        & AssetSelection.keys("be/source_table").downstream(include_self=True)
    )
)


@multi_asset_sensor(
    monitored_assets=assets_to_monitor,
    job=cloud_assets_job,
    minimum_interval_seconds=10,
    default_status=DefaultSensorStatus.STOPPED,
)
def country_outputs_sensor(context):
    """
    Sensor that detects when the individual country outputs are available.
    """
    context.log.info("Country outputs are available")

    ic(context.latest_materialization_records_by_key())
    asset_events = context.latest_materialization_records_by_key()

    have_yielded_run_request = False
    for asset_key, execution_value in asset_events.items():
        context.log.debug(
            ic(f"asset_key: {asset_key}, execution_value: {execution_value}")
        )
        if execution_value:
            asset_to_load = asset_key.to_user_string()

            # Try and get the partition key if it exists, otherwise default to None
            try:
                source_partitions_keys = (
                    execution_value.event_log_entry.dagster_event.event_specific_data.materialization.partition
                )
            except Exception:
                source_partitions_keys = None

            context.log.debug(ic(f"source_partitions_keys: {source_partitions_keys}"))

            # Make a unique key for the run
            if source_partitions_keys:
                run_key = slugify(f"{asset_to_load}-{source_partitions_keys}")
            else:
                run_key = slugify(asset_to_load)

            context.log.debug(ic(f"run_key: {run_key}"))

            # Required, during development, when the pattern for the `run_key`` is changed
            if False:
                for partition_key in context.instance.get_dynamic_partitions(
                    "publishing_partition"
                ):
                    context.instance.delete_dynamic_partition(
                        "publishing_partition", partition_key
                    )

            # Add the run_key to the dynamic partitions if it is not already there
            if run_key not in context.instance.get_dynamic_partitions(
                "publishing_partition"
            ):
                context.instance.add_dynamic_partitions(
                    partitions_def_name="publishing_partition", partition_keys=[run_key]
                )

            for partition_key in context.instance.get_dynamic_partitions(
                "publishing_partition"
            ):
                context.log.debug(ic(f"available partition_key: {partition_key}"))

            have_yielded_run_request = True
            yield RunRequest(
                # TODO: run_key is set to None for now, as it is convenient to enable re-running the sensor
                # whilst debugging. A proper run_key should be included as part of
                # # https://github.com/Urban-Analytics-Technology-Platform/poppusher/issues/38
                # and https://github.com/Urban-Analytics-Technology-Platform/poppusher/issues/39
                run_key=None,
                # run_key=run_key,
                run_config=RunConfig(
                    ops={
                        "upstream_df": CountryAssetConfig(
                            asset_to_load=asset_to_load,
                            partition_to_load=source_partitions_keys,
                        )
                    },
                ),
                partition_key=run_key,
            )
            context.advance_cursor({asset_key: execution_value})

    if not have_yielded_run_request:
        # TODO: Probably don't need to provide this level of detail in the SkipReason
        materialized_asset_key_strs = [
            key.to_user_string() for key, value in asset_events.items() if value
        ]
        not_materialized_asset_key_strs = [
            key.to_user_string() for key, value in asset_events.items() if not value
        ]
        yield SkipReason(
            f"Observed materializations for {materialized_asset_key_strs}, "
            f"but not for {not_materialized_asset_key_strs}"
        )


@asset(
    # todo: this feels like a code smell. (mixing my metaphors)
    # it feels that this structure is duplicated and it ought
    # to be possible to have some reusable structure.
    config_schema={
        "asset_to_load": str,
        "partition_to_load": Noneable(str),
    },
    partitions_def=publishing_partition,
)
def upstream_df(context):
    """
    Returns a GeoDataFrame of raw cartography data, from a country specific pipeline.
    """
    asset_to_load = context.run_config["ops"]["upstream_df"]["config"]["asset_to_load"]

    # partition_to_load might not be present. Default to None if it is not present.
    partition_to_load = context.run_config["ops"]["upstream_df"]["config"].get(
        "partition_to_load", None
    )

    log_msg = f"Beginning conversion of cartography data from '{asset_to_load}' (partition_key={partition_to_load}) into cloud formats."
    context.log.info(log_msg)

    # TODO: This is horribly hacky - I do not like importing stuff in the middle of a module or function
    # However I cannot find another may to load the asset except
    # by using the Definitions object (or creating a RepositoryDefinition which seems to be deprecated)
    # Import the Definitions object here in the function, avoids a circular import error.
    from poppusher import defs as poppusher_defs

    cartography_gdf = poppusher_defs.load_asset_value(
        AssetKey(asset_to_load.split("/")),
        partition_key=partition_to_load,
    )

    if isinstance(cartography_gdf, gpd.GeoDataFrame):
        return Output(cartography_gdf)

    # TODO: This is a temporary solution. We need to handle non-GeoDataFrame types
    err_msg = f"Expected a GeoDataFrame, but got {type(cartography_gdf)}. Handling non-GeoDataFrame types is not yet implemented."
    context.log.error(ic(err_msg))
    raise ValueError(err_msg)


def df_to_bytes(df: gpd.GeoDataFrame, output_type: str) -> bytes:
    tmp = tempfile.NamedTemporaryFile(prefix=str(uuid.uuid4()))
    if output_type.lower() == "parquet":
        df.to_parquet(tmp.name + ".parquet")
    elif output_type.lower() == "flatgeobuf":
        df.to_file(tmp.name + ".flatgeobuf", driver="FlatGeobuf")
    elif output_type.lower() == "geojsonseq":
        df.to_file(tmp.name + ".geojsonseq", driver="GeoJSONSeq")
    else:
        value_error: str = f"'{output_type}' is not currently supported."
        raise ValueError(value_error)
    with Path(tmp.name).open(mode="rb") as f:
        return f.read()


@asset(io_manager_key="publishing_io_manager", partitions_def=publishing_partition)
def parquet(context, upstream_df):  # noqa: ARG001
    return df_to_bytes(upstream_df, "parquet")


@asset(io_manager_key="publishing_io_manager", partitions_def=publishing_partition)
def flatgeobuf(context, upstream_df):  # noqa: ARG001
    return df_to_bytes(upstream_df, "flatgeobuf")


@asset(io_manager_key="publishing_io_manager", partitions_def=publishing_partition)
def geojsonseq(context, upstream_df):  # noqa: ARG001
    return df_to_bytes(upstream_df, "geojsonseq")


# Not working yet - need to figure out questions about how we run docker
# See comments here https://github.com/Urban-Analytics-Technology-Platform/poppusher/pull/68#issue-2205271531
# @asset()
# def generate_pmtiles(context, geojson_seq_path):
#     client = docker.from_env()
#     mount_folder = Path(geojson_seq_path).parent.resolve()
#     container = client.containers.run(
#         "stuartlynn/tippecanoe:latest",
#         "tippecanoe -o tracts.pmtiles tracts.geojsonseq",
#         volumes={mount_folder: {"bind": "/app", "mode": "rw"}},
#         detach=True,
#         remove=True,
#     )
#     output = container.attach(stdout=True, stream=True, logs=True)
#     for line in output:
#         context.log.info(line)



from __future__ import annotations

import os
import warnings
from collections.abc import Sequence
from pathlib import Path

from dagster import ExperimentalWarning

# Has to be placed before imports as otherwise importing other modules will
# trigger ExperimentalWarnings themselves...
if "IGNORE_EXPERIMENTAL_WARNINGS" in os.environ:
    warnings.filterwarnings("ignore", category=ExperimentalWarning)

from poppusher.io_managers.azure import (
    AzureGeneralIOManager,
    AzureGeoIOManager,
    AzureMetadataIOManager,
    AzureMetricsIOManager,
    AzureMetricsMetadataIOManager,
    AzureMetricsPartitionedIOManager,
)
from poppusher.io_managers.local import (
    LocalGeoIOManager,
    LocalMetadataIOManager,
    LocalMetricsIOManager,
    LocalMetricsMetadataIOManager,
    LocalMetricsPartitionedIOManager,
)
from poppusher.utils import StagingDirResource

__version__ = "0.1.0"

__all__ = ["__version__"]


import os

from dagster import (
    AssetsDefinition,
    AssetSelection,
    Definitions,
    PipesSubprocessClient,
    SourceAsset,
    define_asset_job,
    load_assets_from_modules,
    load_assets_from_package_module,
)
from dagster._core.definitions.cacheable_assets import (
    CacheableAssetsDefinition,
)
from dagster._core.definitions.unresolved_asset_job_definition import (
    UnresolvedAssetJobDefinition,
)

from poppusher import azure_test, cloud_outputs
from poppusher.assets import countries

PROD = os.getenv("ENV") == "prod"

all_assets: Sequence[AssetsDefinition | SourceAsset | CacheableAssetsDefinition] = [
    *[
        asset
        for (module, name) in countries
        for asset in load_assets_from_package_module(module, group_name=name)
    ],
    *load_assets_from_package_module(cloud_outputs, group_name="cloud_outputs"),
    *(load_assets_from_modules([azure_test], group_name="azure_test") if PROD else []),
]

jobs: list[UnresolvedAssetJobDefinition] = [
    define_asset_job(
        name=f"job_{country_name}",
        selection=AssetSelection.groups(country_name),
        description=f"Downloads data for country '{country_name}'.",
    )
    for (_, country_name) in countries
]


def resources_by_env():
    return (
        {
            "metadata_io_manager": AzureMetadataIOManager(),
            "geometry_io_manager": AzureGeoIOManager(),
            "metrics_io_manager": AzureMetricsIOManager(),
            "azure_general_io_manager": AzureGeneralIOManager(".bin"),
            "metrics_partitioned_io_manager": AzureMetricsPartitionedIOManager(),
            "metrics_metadata_io_manager": AzureMetricsMetadataIOManager(),
        }
        if PROD
        else {
            "metadata_io_manager": LocalMetadataIOManager(),
            "geometry_io_manager": LocalGeoIOManager(),
            "metrics_io_manager": LocalMetricsIOManager(),
            "metrics_partitioned_io_manager": LocalMetricsPartitionedIOManager(),
            "metrics_metadata_io_manager": LocalMetricsMetadataIOManager(),
        }
    )


resources = {
    "pipes_subprocess_client": PipesSubprocessClient(),
    "staging_res": StagingDirResource(
        staging_dir=str(Path(__file__).parent.joinpath("staging_dir").resolve())
    ),
} | resources_by_env()


defs: Definitions = Definitions(
    assets=all_assets,
    schedules=[],
    sensors=[
        cloud_outputs.metadata_sensor,
        cloud_outputs.geometry_sensor,
        cloud_outputs.metrics_sensor,
        cloud_outputs.metrics_metadata_sensor,
        # TODO: commented out until implemented for partitioned assets
        # cloud_outputs.metrics_partitioned_sensor,
    ],
    resources=resources,
    jobs=jobs,
)



#!/usr/bin/python3
from __future__ import annotations

import urllib.parse as urlparse
from collections.abc import Callable
from dataclasses import dataclass
from datetime import date
from functools import reduce
from pathlib import Path
from typing import ClassVar

import geopandas as gpd
import pandas as pd
import zipfile_deflate64 as zipfile
from dagster import (
    MetadataValue,
    asset,
)
from icecream import ic

from poppusher.assets.country import Country
from poppusher.assets.gb_sct.utils import HEADERS, download_file
from poppusher.cloud_outputs import (
    GeometryOutput,
    MetricsOutput,
    send_to_geometry_sensor,
)
from poppusher.metadata import (
    COL,
    CountryMetadata,
    DataPublisher,
    GeometryMetadata,
    MetricMetadata,
    SourceDataRelease,
    metadata_to_dataframe,
)
from poppusher.utils import add_metadata, markdown_from_plot

# Currently including only releases matching tables included
GENERAL_METHODS_URL = "https://www.scotlandscensus.gov.uk/media/jx2lz54n/scotland-s_census_2011_general_report.pdf"
CENSUS_REFERENCE_DATE = date(2011, 3, 27)
CENSUS_COLLECTION_DATE = date(2011, 3, 27)
CENSUS_EXPECT_NEXT_UPDATE = date(2022, 1, 1)


# Source releases for 2011:
# '3A','3I', '2A', '3C', '3D', '3E', '3L', '3K', '3N',
# '3B', '3J', '3M', '3G', '3H', '2C', '2B', '2D',
# Others:
# '2001 Census',
# 'nan', '75+', 'Daytime Tables',
# '1991 Census', '1992 Census',
SOURCE_DATA_RELEASES: dict[str, SourceDataRelease] = {
    "3A": SourceDataRelease(
        name="Census 2011: Release 3A",
        date_published=date(2014, 2, 27),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3a",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3I": SourceDataRelease(
        name="Census 2011: Release 3I",
        date_published=date(2014, 9, 24),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3i",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "2A": SourceDataRelease(
        name="Census 2011: Release 2A",
        date_published=date(2013, 9, 26),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2013/census-2011-release-2a",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3C": SourceDataRelease(
        name="Census 2011: Release 3C",
        date_published=date(2014, 4, 9),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-releases-2d-and-3c",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3D": SourceDataRelease(
        name="Census 2011: Release 3D",
        date_published=date(2014, 5, 15),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3d",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3E": SourceDataRelease(
        name="Census 2011: Release 3E",
        date_published=date(2014, 6, 4),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3e",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3F": SourceDataRelease(
        name="Census 2011: Release 3F",
        date_published=date(2014, 6, 25),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-release-3f",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3L": SourceDataRelease(
        name="Census 2011: Release 3L",
        date_published=date(2014, 11, 27),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3l",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3K": SourceDataRelease(
        name="Census 2011: Release 3K",
        date_published=date(2014, 11, 6),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3k",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3N": SourceDataRelease(
        name="Census 2011: Release 3N",
        date_published=date(2015, 1, 29),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2015/census-2011-release-3n",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3B": SourceDataRelease(
        name="Census 2011: Release 3B",
        date_published=date(2014, 3, 19),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-detailed-characteristics-on-ethnicity-identity-language-and-religion-in-scotland-%E2%80%93-release-3b",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3J": SourceDataRelease(
        name="Census 2011: Release 3J",
        date_published=date(2014, 10, 16),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3j",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3M": SourceDataRelease(
        name="Census 2011: Release 3M",
        date_published=date(2014, 12, 18),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3m",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3G": SourceDataRelease(
        name="Census 2011: Release 3G",
        date_published=date(2014, 7, 23),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3g",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "3H": SourceDataRelease(
        name="Census 2011: Release 3H",
        date_published=date(2014, 8, 13),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-release-3h",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "2C": SourceDataRelease(
        name="Census 2011: Release 2C",
        date_published=date(2013, 12, 18),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2013/census-2011-release-2c",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "2B": SourceDataRelease(
        name="Census 2011: Release 2B",
        date_published=date(2013, 11, 14),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2013/census-2011-release-2b",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "2D": SourceDataRelease(
        name="Census 2011: Release 2D",
        date_published=date(2014, 4, 9),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="https://www.nrscotland.gov.uk/news/2014/census-2011-releases-2d-and-3c",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
    "75+": SourceDataRelease(
        name="Census 2011: 75+",
        # TODO: unable to find published date for 75+ release
        date_published=date(2014, 1, 1),
        reference_period_start=CENSUS_REFERENCE_DATE,
        reference_period_end=CENSUS_REFERENCE_DATE,
        collection_period_start=CENSUS_COLLECTION_DATE,
        collection_period_end=CENSUS_COLLECTION_DATE,
        expect_next_update=CENSUS_EXPECT_NEXT_UPDATE,
        url="TBD",
        data_publisher_id="TBD",
        description="TBD",
        geometry_metadata_id="TBD",
    ),
}


# TODO: remove ones no longer used
URL = "https://www.scotlandscensus.gov.uk/ods-web/download/getDownloadFile.html"
URL_LOOKUP = (
    "https://www.nrscotland.gov.uk/files//geography/2011-census/OA_DZ_IZ_2011.xlsx"
)
URL_SHAPEFILE = "https://borders.ukdataservice.ac.uk/ukborders/easy_download/prebuilt/shape/infuse_oa_lyr_2011.zip"
URL_CATALOG = (
    "https://www.scotlandscensus.gov.uk/media/kqcmo4ge/census-table-index-2011.xlsm"
)


CENSUS_TABLE_DATA_SOURCES = [
    {
        "source": "Council Area blk",
        # "resolution": "LAD",
        "resolution": "CouncilArea2011",
        "url": "https://www.scotlandscensus.gov.uk/media/hjmd0oqr/council-area-blk.zip",
    },
    {
        "source": "SNS Data Zone 2011 blk",
        # "resolution": "LSOA11",
        "resolution": "DataZone2011",
        "url": "https://nrscensusprodumb.blob.core.windows.net/downloads/"
        + urlparse.quote("SNS Data Zone 2011 blk")
        + ".zip",
    },
    {
        "source": "Output Area blk",
        # "resolution": "OA11",
        "resolution": "OutputArea2011",
        "url": "https://nrscensusprodumb.blob.core.windows.net/downloads/"
        + urlparse.quote("Output Area blk")
        + ".zip",
    },
]


@dataclass
class ScotlandGeometryLevel:
    level: str
    hxl_tag: str
    geo_id_column: str
    name_columns: dict[str, str]  # keys = language codes, values = column names
    url: str
    lookup_url: str | None
    lookup_sheet: str | None
    left_on: str | None
    right_on: str | None


SCOTLAND_GEO_LEVELS = {
    "OutputArea2011": ScotlandGeometryLevel(
        level="OutputArea2011",
        hxl_tag="TBD",
        geo_id_column="OA_CODE",
        name_columns={"eng": "OutputArea2011Name"},
        url="https://www.nrscotland.gov.uk/files/geography/output-area-2011-eor.zip",
        lookup_url=None,
        lookup_sheet=None,
        left_on="OA_CODE",
        right_on="OutputArea2011Code",
    ),
    # LSOA11
    "DataZone2011": ScotlandGeometryLevel(
        level="DataZone2011",
        hxl_tag="TBD",
        geo_id_column="DataZone",
        name_columns={"eng": "Name"},
        url="https://maps.gov.scot/ATOM/shapefiles/SG_DataZoneBdry_2011.zip",
        lookup_url=None,
        lookup_sheet=None,
        left_on="DataZone",
        right_on="DataZone2011Code",
    ),
    # "MSOA11": ScotlandGeometryLevel(
    #     level="OA11",
    #     hxl_tag="TBD",
    #     geo_id_column="OA_CODE",
    #     name_columns={"eng": "OA_CODE"},
    #     # url=URL_SHAPEFILE,
    #     url="https://www.nrscotland.gov.uk/files/geography/output-area-2011-eor.zip",
    #     lookup_url=None,
    #     lookup_sheet=None,
    #     left_on=None,
    #     right_on=None,
    # ),
    # LAD
    "CouncilArea2011": ScotlandGeometryLevel(
        level="CouncilArea2011",
        hxl_tag="TBD",
        geo_id_column="CouncilArea2011Code",
        name_columns={"eng": "CouncilArea2011Name"},
        url="https://maps.gov.scot/ATOM/shapefiles/SG_DataZoneBdry_2011.zip",
        lookup_url=None,
        lookup_sheet=None,
        left_on="DataZone",
        right_on="DataZone2011Code",
    ),
}


# TODO: identify better tempfile option
# Use temporary directory
# CACHE_DIR = tempfile.mkdtemp()
CACHE_DIR = "./cache"


@dataclass
class DerivedColumn:
    hxltag: str
    filter_func: Callable[[pd.DataFrame], pd.DataFrame]
    output_column_name: str
    human_readable_name: str


@dataclass
class SourceTable:
    hxltag: str
    geo_level: str
    geo_column: str
    source_column: str


# Config for each partition to be derived
AGE_CODE = "`Age Category`"
SEX_LABEL = "`Sex Label`"
INFANTS = ["0 to 4"]
CHILDREN_5_TO_17 = ["5 to 9", "10 to 11", "12 to 14" "15", "16 to 17"]
CHILDREN = ["0 to 4", "5 to 9", "10 to 11", "12 to 14" "15", "16 to 17"]
ADULTS = ["18 to 19"] + [f"{i} to {i+4}" for i in range(20, 91, 5)] + ["95 and over"]
PEOPLE = ["All people"]
DERIVED_COLUMNS = [
    DerivedColumn(
        hxltag="#population+children+age5_17",
        filter_func=lambda df: df.query(f"{AGE_CODE} in @CHILDREN_5_TO_17"),
        output_column_name="children_5_17",
        human_readable_name="Children aged 5 to 17",
    ),
    DerivedColumn(
        hxltag="#population+infants+age0_4",
        filter_func=lambda df: df.query(f"{AGE_CODE} in @INFANTS"),
        output_column_name="infants_0_4",
        human_readable_name="Infants aged 0 to 4",
    ),
    DerivedColumn(
        hxltag="#population+children+age0_17",
        filter_func=lambda df: df.query(f"{AGE_CODE} in @CHILDREN"),
        output_column_name="children_0_17",
        human_readable_name="Children aged 0 to 17",
    ),
    DerivedColumn(
        hxltag="#population+adults+f",
        filter_func=lambda df: df.query(
            f"{AGE_CODE} in @ADULTS and {SEX_LABEL} == 'Female'"
        ),
        output_column_name="adults_f",
        human_readable_name="Female adults",
    ),
    DerivedColumn(
        hxltag="#population+adults+m",
        filter_func=lambda df: df.query(
            f"{AGE_CODE} in @ADULTS and {SEX_LABEL} == 'Male'"
        ),
        output_column_name="adults_m",
        human_readable_name="Male adults",
    ),
    DerivedColumn(
        hxltag="#population+adults",
        filter_func=lambda df: df.query(f"{AGE_CODE} in @ADULTS"),
        output_column_name="adults",
        human_readable_name="Adults",
    ),
    DerivedColumn(
        hxltag="#population+ind",
        filter_func=lambda df: df.query(f"{AGE_CODE} in @PEOPLE"),
        output_column_name="individuals",
        human_readable_name="Total individuals",
    ),
]

# For a subset:
# TABLES_TO_PROCESS: list[str] = [
#     "QS103SC",
#     "QS104SC",
#     "KS201SC",
#     "DC1117SC",
#     "DC2101SC",
#     "DC6206SC",
#     "LC1117SC",
# ]

# For all available:
TABLES_TO_PROCESS = None

DERIVED_COLUMN_SPECIFICATIONS: dict[str, list[DerivedColumn]] = {
    "2011/OutputArea2011/LC1117SC": DERIVED_COLUMNS,
    "2011/DataZone2011/LC1117SC": DERIVED_COLUMNS,
    "2011/CouncilArea2011/LC1117SC": DERIVED_COLUMNS,
}


def get_source_data_release(geo_level: str, cenesus_release: str) -> str:
    return geo_level + "_" + cenesus_release


class Scotland(Country):
    country_metadata: ClassVar[CountryMetadata] = CountryMetadata(
        name_short_en="Scotland",
        name_official="Scotland",
        iso3="GBR",
        iso2="GB",
        iso3166_2="GB-SCT",
    )
    geo_levels: ClassVar[list[str]] = list(SCOTLAND_GEO_LEVELS.keys())
    tables_to_process: list[str] | None = TABLES_TO_PROCESS
    allow_missing_derived_metrics: ClassVar[bool] = True

    def __init__(self):
        super().__init__()

        # Make temp directory
        Path(CACHE_DIR).mkdir(parents=True, exist_ok=True)

    def _catalog(self, context) -> pd.DataFrame:
        """Creates a catalog of the individual census tables from all data sources."""

        def source_to_zip(source_name: str, url: str) -> Path:
            """Downloads if necessary and returns the name of the locally cached zip file
            of the source data (replacing spaces with _)"""
            file_name = Path(CACHE_DIR) / (source_name.replace(" ", "_") + ".zip")
            return download_file(CACHE_DIR, url, file_name)

        def get_table_name(file_name: str) -> str:
            return file_name.rsplit(".csv")[0]

        def get_table_metadata(
            catalog_reference: pd.DataFrame, table_name: str
        ) -> dict[str, str]:
            """Returns a dict of table metadata for a given table name."""
            rows = catalog_reference.loc[
                catalog_reference.loc[:, "table_name"].eq(table_name)
            ]
            census_release = rows.loc[:, "census_release"].unique()[0]
            description = rows.loc[:, "description"].unique()[0]
            population_coverage = rows.loc[:, "population_coverage"].unique()[0]
            variables = ", ".join(rows.loc[:, "variable"].astype(str).to_list())
            catalog_resolution = rows.loc[:, "catalog_resolution"].unique()[0]
            year = int(rows.loc[:, "year"].unique()[0])
            return {
                "census_release": census_release,
                "description": description,
                "population_coverage": population_coverage,
                "variables": variables,
                "catalog_resolution": catalog_resolution,
                "year": str(year),
                "human_readable_name": f"{description} ({population_coverage})",
            }

        # Download catalog reference
        catalog_reference = pd.read_excel(
            URL_CATALOG,
            sheet_name=None,
            header=None,
            storage_options={"User-Agent": "Mozilla/5.0"},
        )["Index"].rename(
            columns={
                0: "census_release",
                1: "table_name",
                2: "description",
                3: "population_coverage",
                4: "variable",
                5: "catalog_resolution",
                6: "year",
                7: "additional_url",
                8: "population_coverage_and_variable",
            }
        )
        # Remove all keys
        self.remove_all_partition_keys(context)

        records = []
        for data_source in CENSUS_TABLE_DATA_SOURCES:
            resolution = data_source["resolution"]
            source = data_source["source"]
            url = data_source["url"]
            zip_file_name = source_to_zip(source, url)
            with zipfile.ZipFile(zip_file_name) as zip_ref:
                for file_name in zip_ref.namelist():
                    # Get table name
                    table_name = get_table_name(file_name)

                    # Skip bulk output files and missing tables from catalog_reference
                    if (
                        "bulk_output" in file_name.lower()
                        or catalog_reference.loc[:, "table_name"].ne(table_name).all()
                    ):
                        continue

                    # Get table metadata
                    table_metadata = get_table_metadata(catalog_reference, table_name)

                    # Get source release metadata if available
                    source_data_release = SOURCE_DATA_RELEASES.get(
                        table_metadata["census_release"], None
                    )
                    source_data_release_id = (
                        None if source_data_release is None else source_data_release.id
                    )

                    # Skip if not required
                    if (
                        self.tables_to_process is not None
                        and table_name not in self.tables_to_process
                    ):
                        continue

                    # Remove failing case (no data in census table):
                    #   "2011/DataZone2011/QS421SC"
                    if table_name == "QS421SC" and resolution == "DataZone2011":
                        continue

                    # Fix case with missing data for release
                    if resolution == "CouncilArea2011" and table_name == "DC6102SC":
                        table_metadata["census_release"] = "3I"

                    # Create a record for each census table use same keys as MetricMetadata
                    # where possible since this makes it simpler to populate derived
                    # metrics downstream
                    record = {
                        "resolution": resolution,
                        "catalog_resolution": table_metadata["catalog_resolution"],
                        "source": source,
                        "url": url,
                        "file_name": Path(source) / file_name,
                        "table_name": table_name,
                        "year": table_metadata["year"],
                        # Use constructed name of description and coverage
                        "human_readable_name": table_metadata["human_readable_name"],
                        "source_metric_id": None,
                        # Use catalog_reference description
                        "description": table_metadata["description"],
                        "hxl_tag": None,
                        "metric_parquet_file_url": None,
                        "parquet_column_name": None,
                        "parquet_margin_of_error_column": None,
                        "parquet_margin_of_error_file": None,
                        "potential_denominator_ids": None,
                        "parent_metric_id": None,
                        # TODO: check this is not an ID but a name
                        "source_data_release_id": source_data_release_id,
                        "census_release": table_metadata["census_release"],
                        "source_download_url": url,
                        # TODO: what should this be?
                        "source_archive_file_path": None,
                        "source_documentation_url": URL_CATALOG,
                    }
                    context.log.debug(record)
                    records.append(record)
                    zip_ref.extract(file_name, Path(CACHE_DIR) / source)

        # Create a dynamic partition for the datasets listed in the catalog
        catalog_df: pd.DataFrame = pd.DataFrame.from_records(records)
        catalog_df["partition_key"] = (
            catalog_df[["year", "resolution", "table_name"]]
            .astype(str)
            .agg(lambda s: "/".join(s).rsplit(".")[0], axis=1)
        )

        # TODO: add filter for prod vs. dev mode
        self.add_partition_keys(context, catalog_df["partition_key"].to_list())
        context.add_output_metadata(
            metadata={
                "num_records": len(catalog_df),
                "ignored_datasets": "",
                "columns": MetadataValue.md(
                    "\n".join([f"- '`{col}`'" for col in catalog_df.columns.to_list()])
                ),
                "columns_types": MetadataValue.md(catalog_df.dtypes.to_markdown()),
                "preview": MetadataValue.md(catalog_df.to_markdown()),
            }
        )
        return catalog_df

    def _country_metadata(self, _context) -> CountryMetadata:
        return self.country_metadata

    def _data_publisher(
        self, _context, country_metdata: CountryMetadata
    ) -> DataPublisher:
        return DataPublisher(
            name="National Records of Scotland",
            url="https://www.nrscotland.gov.uk/",
            description="National Records of Scotland (NRS) is a Non-Ministerial Department of "
            "the Scottish Government. Our purpose is to collect, preserve and "
            "produce information about Scotland's people and history and make it "
            "available to inform current and future generations.",
            countries_of_interest=[country_metdata.id],
        )

    def create_lookup(self):
        @asset(key_prefix=self.key_prefix)
        def lookup(context):
            url = "https://www.nrscotland.gov.uk/files/geography/2011-census/geog-2011-cen-supp-info-oldoa-newoa-lookup.xls"
            df_oa_to_council = (
                pd.read_excel(url, sheet_name="2011OA_Lookup", storage_options=HEADERS)
                .iloc[:-2]
                .loc[:, ["OutputArea2011Code", "CouncilArea2011Code"]]
            )
            url = "https://www.nrscotland.gov.uk/files//geography/2011-census/OA_DZ_IZ_2011.xlsx"
            df_oa_to_dz_iz = pd.read_excel(
                url, sheet_name="OA_DZ_IZ_2011 Lookup", storage_options=HEADERS
            )
            df_dz_nm = pd.read_excel(
                url, sheet_name="DataZone2011Lookup", storage_options=HEADERS
            )
            df_iz_nm = pd.read_excel(
                url, sheet_name="IntermediateZone2011Lookup", storage_options=HEADERS
            )
            combined = (
                df_oa_to_council.merge(df_oa_to_dz_iz, on=["OutputArea2011Code"])
                .merge(df_dz_nm, on=["DataZone2011Code"])
                .merge(df_iz_nm, on=["IntermediateZone2011Code"])
            )
            combined["OutputArea2011Name"] = combined["OutputArea2011Code"].copy()
            df_council_name = pd.read_excel(
                "https://www.nrscotland.gov.uk/files//geography/2011-census/oa2011-to-hba2014.xls",
                sheet_name="HealthBoard2014_Council2011",
                storage_options=HEADERS,
            )
            combined = combined.merge(
                df_council_name[["CouncilArea2011Code", "NRSCouncilAreaName"]],
                on="CouncilArea2011Code",
            ).rename(columns={"NRSCouncilAreaName": "CouncilArea2011Name"})
            context.add_output_metadata(
                metadata={
                    "lookup_shape": f"{combined.shape[0]} rows x {combined.shape[1]} columns",
                    "lookup_preview": MetadataValue.md(combined.head().to_markdown()),
                },
            )
            return combined

        return lookup

    def create_geometry(self):
        """
        Creates an asset providing a list of geometries, metadata and names
        at different resolutions.
        """

        @send_to_geometry_sensor
        @asset(key_prefix=self.key_prefix)
        def geometry(context, lookup: pd.DataFrame) -> list[GeometryOutput]:
            """List of geometries, metadata and names at different resolutions."""
            geometries_to_return = []
            for level_details in SCOTLAND_GEO_LEVELS.values():
                # TODO: get correct values
                geometry_metadata = GeometryMetadata(
                    country_metadata=self.country_metadata,
                    validity_period_start=CENSUS_COLLECTION_DATE,
                    validity_period_end=CENSUS_COLLECTION_DATE,
                    level=level_details.level,
                    hxl_tag=level_details.hxl_tag,
                )
                file_name = download_file(CACHE_DIR, level_details.url)
                region_geometries_raw: gpd.GeoDataFrame = gpd.read_file(
                    f"zip://{file_name}"
                )
                context.log.debug(ic(region_geometries_raw.head()))
                context.log.debug(ic(region_geometries_raw.columns))
                context.log.debug(ic(lookup.columns))
                region_geometries_merge = region_geometries_raw.merge(
                    lookup,
                    left_on=level_details.left_on,
                    right_on=level_details.right_on,
                )

                region_geometries_merge = region_geometries_merge.dissolve(
                    by=level_details.geo_id_column
                ).reset_index()

                context.log.debug(ic(region_geometries_merge.head()))
                context.log.debug(ic(region_geometries_merge.columns))
                region_geometries = region_geometries_merge.rename(
                    columns={level_details.geo_id_column: COL.GEO_ID.value}
                ).loc[:, ["geometry", COL.GEO_ID.value]]

                region_names = (
                    region_geometries_merge.rename(
                        columns={
                            level_details.geo_id_column: COL.GEO_ID.value,
                        }
                        | {
                            value: key
                            for key, value in level_details.name_columns.items()
                        }
                    )
                    .loc[
                        :, [COL.GEO_ID.value, *list(level_details.name_columns.keys())]
                    ]
                    .drop_duplicates()
                )
                geometries_to_return.append(
                    GeometryOutput(
                        metadata=geometry_metadata,
                        gdf=region_geometries,
                        names_df=region_names,
                    )
                )

            # Add output metadata
            geo: GeometryOutput = geometries_to_return[0]
            first_metadata, first_gdf, first_names = geo.metadata, geo.gdf, geo.names_df
            first_joined_gdf = first_gdf.merge(first_names, on=COL.GEO_ID.value)
            ax = first_joined_gdf.plot(column="eng", legend=False)
            ax.set_title(f"Scotland 2011 {first_metadata.level}")
            md_plot = markdown_from_plot()
            context.add_output_metadata(
                metadata={
                    "all_geom_levels": MetadataValue.md(
                        ",".join([geo.metadata.level for geo in geometries_to_return])
                    ),
                    "first_geometry_plot": MetadataValue.md(md_plot),
                    "first_names_preview": MetadataValue.md(
                        first_names.head().to_markdown()
                    ),
                }
            )

            return geometries_to_return

        return geometry

    def _geometry(self, context):
        # Not required as geometry overridden
        pass

    def _source_data_releases(
        self,
        _context,
        geometry: list[GeometryOutput],
        data_publisher: DataPublisher,
        # TODO: consider version without inputs so only output type specified
        # **kwargs,
    ) -> dict[str, SourceDataRelease]:
        source_data_releases = {}
        for geo in geometry:
            for (
                source_data_release_id,
                source_data_release,
            ) in SOURCE_DATA_RELEASES.items():
                source_data_release_new: SourceDataRelease = SourceDataRelease(
                    name=source_data_release.name,
                    date_published=source_data_release.date_published,
                    reference_period_start=source_data_release.collection_period_start,
                    reference_period_end=source_data_release.reference_period_end,
                    collection_period_start=source_data_release.collection_period_start,
                    collection_period_end=source_data_release.collection_period_end,
                    expect_next_update=source_data_release.expect_next_update,
                    url=source_data_release.url,
                    data_publisher_id=data_publisher.id,
                    description=source_data_release.description,
                    geometry_metadata_id=geo.metadata.id,
                )
                combined_level_and_release_id = get_source_data_release(
                    geo.metadata.level, source_data_release_id
                )
                source_data_releases[
                    combined_level_and_release_id
                ] = source_data_release_new
        return source_data_releases

    @staticmethod
    def get_table(context, table_details) -> pd.DataFrame:
        table_df = pd.read_csv(Path(CACHE_DIR) / table_details["file_name"].iloc[0])
        add_metadata(context, table_df, table_details["partition_key"].iloc[0])
        return table_df

    def _census_tables(self, context, catalog: pd.DataFrame) -> pd.DataFrame:
        """Creates individual census tables as dataframe."""
        partition_key = context.asset_partition_key_for_output()
        context.log.info(partition_key)
        table_details = catalog.loc[catalog["partition_key"].isin([partition_key])]
        context.log.info(table_details)
        return self.get_table(context, table_details)

    @staticmethod
    def census_table_metadata(
        catalog_row: dict[str, str],
        source_table: SourceTable,
        source_data_releases: dict[str, SourceDataRelease],
    ) -> MetricMetadata:
        return MetricMetadata(
            human_readable_name=catalog_row["human_readable_name"],
            source_download_url=catalog_row["source_download_url"],
            source_archive_file_path=catalog_row["source_archive_file_path"],
            source_documentation_url=catalog_row["source_documentation_url"],
            source_data_release_id=source_data_releases[
                get_source_data_release(
                    source_table.geo_level, catalog_row["census_release"]
                )
            ].id,
            # TODO - this is a placeholder
            parent_metric_id="unknown_at_this_stage",
            potential_denominator_ids=None,
            parquet_margin_of_error_file=None,
            parquet_margin_of_error_column=None,
            parquet_column_name=source_table.source_column,
            # TODO - this is a placeholder
            metric_parquet_path="unknown_at_this_stage",
            hxl_tag=source_table.hxltag,
            description=catalog_row["description"],
            source_metric_id=source_table.hxltag,
        )

    def _source_metric_metadata(
        self,
        context,
        catalog: pd.DataFrame,
        source_data_releases: dict[str, SourceDataRelease],
    ) -> MetricMetadata:
        partition_key = context.partition_key
        catalog_row = catalog[catalog["partition_key"] == partition_key].to_dict(
            orient="records"
        )[0]

        geo_level = partition_key.split("/")[1]
        source_table = SourceTable(
            # TODO: how programmatically do this
            hxltag="TBD",
            geo_level=geo_level,
            geo_column=SCOTLAND_GEO_LEVELS[geo_level].geo_id_column,
            # TODO: update this
            source_column="Count",
        )

        return self.census_table_metadata(
            catalog_row,
            source_table,
            source_data_releases,
        )

    def _derived_metrics(
        self,
        context,
        census_tables: pd.DataFrame,
        source_metric_metadata: MetricMetadata,
    ) -> MetricsOutput:
        try:
            SEP = "__"
            partition_key = context.partition_key
            source_mmd = source_metric_metadata
            parquet_file_name = (
                f"{self.key_prefix}/metrics/"
                f"{''.join(c for c in partition_key if c.isalnum()) + '.parquet'}"
            )
            derived_metrics, derived_mmd = [], []

            # If derived metrics
            try:
                metric_specs = DERIVED_COLUMN_SPECIFICATIONS[partition_key]

                def reshape(df_to_reshape: pd.DataFrame) -> pd.DataFrame:
                    df_to_reshape = df_to_reshape.rename(
                        columns={
                            "Unnamed: 0": COL.GEO_ID.value,
                            "Unnamed: 1": "Age Category",
                        }
                    ).drop(columns=["All people"])
                    df_to_reshape = df_to_reshape.melt(
                        [COL.GEO_ID.value, "Age Category"],
                        var_name="Sex Label",
                        value_name="Count",
                    )
                    df_to_reshape["Sex Label"] = df_to_reshape["Sex Label"].map(
                        {"Males": "Male", "Females": "Female"}
                    )
                    return df_to_reshape

                census_tables_for_derived_metrics = reshape(census_tables)
                source_column = source_mmd.parquet_column_name
                for metric_spec in metric_specs:
                    new_table = (
                        census_tables_for_derived_metrics.pipe(metric_spec.filter_func)
                        .groupby(by=COL.GEO_ID.value, as_index=True)
                        .sum()
                        .rename(columns={source_column: metric_spec.output_column_name})
                        .filter(
                            items=[COL.GEO_ID.value, metric_spec.output_column_name]
                        )
                    )
                    derived_metrics.append(new_table)
                    new_mmd = source_mmd.model_copy(deep=True)
                    new_mmd.parent_metric_id = source_mmd.source_metric_id
                    new_mmd.metric_parquet_path = parquet_file_name
                    new_mmd.hxl_tag = metric_spec.hxltag
                    new_mmd.parquet_column_name = metric_spec.output_column_name
                    new_mmd.human_readable_name = metric_spec.human_readable_name
                    derived_mmd.append(new_mmd)
            except KeyError:
                # No extra derived metrics specified for this partition -- only use
                # those from pivoted data
                pass

            # Batch
            def make_pivot(df: pd.DataFrame) -> pd.DataFrame:
                # TODO: reshape based on Unnamed: 1 to Unnamed N
                pivot_cols = [
                    col
                    for col in df.columns
                    if col != "Unnamed: 0" and col.startswith("Unnamed: ")
                ]
                pivot = df.pivot_table(
                    index="Unnamed: 0", columns=pivot_cols, aggfunc="sum"
                )

                # FLattent multi-index
                if isinstance(pivot.columns, pd.MultiIndex):
                    pivot.columns = [
                        SEP.join(list(map(str, col))).strip()
                        for col in pivot.columns.to_numpy()
                    ]
                # Ensure columns are string
                else:
                    pivot.columns = [
                        str(col).strip() for col in pivot.columns.to_numpy()
                    ]

                pivot.index = pivot.index.rename(COL.GEO_ID.value)

                return pivot

            new_table = make_pivot(census_tables)

            # Split for description of metrics
            exceptions = {
                "Age by single year": ["Age by single year"],
                "National Statistics Socio-economic Classification (NS-SeC) by ethnic group by sex by age": [
                    "Ethnic group",
                    "Sex and Age",
                    "National Statistics Socio-economic Classification (NS-SeC)",
                ],
                # 2011/CouncilArea2011/DC1104SC
                "Residence type by sex by age": ["Residence type and Sex", "Age"],
                # 2011/CouncilArea2011/DC1106SC
                "Schoolchildren and full-time students living away from home during term time by sex by age": [
                    "Schoolchildren and full-time students living away from home during term time and Sex",
                    "Age",
                ],
                # 2011/CouncilArea2011/DC1112SC
                "Dependent children by household type by sex by age": [
                    "Dependent children by household type",
                    "Sex",
                    "Age",
                ],
            }
            if source_mmd.description not in exceptions:
                split = source_mmd.description.split(" by ")[::-1]
            else:
                split = exceptions[source_mmd.description]
            out_cols = [
                "".join(x for x in col.title() if not x.isspace()) for col in split
            ]
            context.log.debug(ic(out_cols))
            ic(new_table.columns)

            for metric_col in new_table.columns:
                metric_df = new_table.loc[:, metric_col].to_frame()
                ic(metric_df)
                derived_metrics.append(metric_df)
                new_mmd = source_mmd.copy()
                new_mmd.parent_metric_id = source_mmd.source_metric_id
                new_mmd.metric_parquet_path = parquet_file_name

                # TODO: fix automating the hxltag
                key_val = dict(zip(out_cols, metric_col.split(SEP), strict=True))

                def gen_hxltag(kv: dict[str, str]) -> str:
                    out = ["#population"]
                    for key, value in kv.items():
                        out += [
                            "".join(c for c in key if c.isalnum())
                            + "_"
                            + "".join(c for c in value if c.isalnum())
                        ]
                    return "+".join(out)

                new_mmd.hxl_tag = gen_hxltag(key_val)
                new_mmd.parquet_column_name = metric_col
                context.log.debug(ic(key_val))
                # TODO: Update after fixing hxltag
                new_mmd.human_readable_name = "; ".join(
                    [
                        f"Variable: '{key}'; Value: '{value}'"
                        for key, value in key_val.items()
                    ]
                )
                derived_mmd.append(new_mmd)

            joined_metrics = reduce(
                lambda left, right: left.merge(
                    right, on=COL.GEO_ID.value, how="inner", validate="one_to_one"
                ),
                derived_metrics,
            )

            def make_int(maybe_non_int_df: pd.DataFrame) -> pd.DataFrame:
                for col in maybe_non_int_df:
                    if maybe_non_int_df[col].dtype == "object":
                        maybe_non_int_df[col] = (
                            maybe_non_int_df[col]
                            .str.replace(",", "")
                            .str.replace("-", "0")
                            .fillna("0")
                            .astype(int)
                        )
                return maybe_non_int_df

            # Fix format
            joined_metrics = make_int(joined_metrics)

            # Filter out whole country Scotland
            joined_metrics = joined_metrics.loc[
                ~joined_metrics.index.isin(["S92000003"])
            ]

            context.add_output_metadata(
                metadata={
                    "metadata_preview": MetadataValue.md(
                        metadata_to_dataframe(derived_mmd).head().to_markdown()
                    ),
                    "metrics_shape": f"{joined_metrics.shape[0]} rows x {joined_metrics.shape[1]} columns",
                    "metrics_preview": MetadataValue.md(
                        joined_metrics.head().to_markdown()
                    ),
                },
            )

        except Exception as err:
            err_msg = (
                f"Failed to automatically derive metrics for '{partition_key}' with "
                f"error: {err}"
            )
            context.log.error(err_msg)
            if self.allow_missing_derived_metrics:
                return MetricsOutput(metadata=[], metrics=pd.DataFrame())

        return MetricsOutput(metadata=derived_mmd, metrics=joined_metrics)


# Create assets
scotland = Scotland()
country_metadata = scotland.create_country_metadata()
data_publisher = scotland.create_data_publisher()
lookup = scotland.create_lookup()
geometry = scotland.create_geometry()
source_data_releases = scotland.create_source_data_releases()
catalog = scotland.create_catalog()
census_tables = scotland.create_census_tables()
source_metric_metadata = scotland.create_source_metric_metadata()
derived_metrics = scotland.create_derived_metrics()
# Publish all partitions
metrics = scotland.create_metrics(None)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import ClassVar

import pandas as pd
from dagster import (
    AssetIn,
    DynamicPartitionsDefinition,
    SpecificPartitionsPartitionMapping,
    asset,
)

from poppusher.cloud_outputs import (
    GeometryOutput,
    MetricsOutput,
    send_to_geometry_sensor,
    send_to_metadata_sensor,
    send_to_metrics_sensor,
)
from poppusher.metadata import (
    CountryMetadata,
    DataPublisher,
    MetricMetadata,
    SourceDataRelease,
)


class Country(ABC):
    """
    A general class that can be implemented for a given country providing asset
    factories and abstract methods to provide a template for a given country.

    Attributes:
        key_prefix (str): the prefix for the asset keys (e.g. "be" for Belgium)
        dataset_node_partition (DynamicPartitionsDefinition): a dynamic partitions
            definition populated at runtime with a partition per census table.

    """

    country_metadata: ClassVar[CountryMetadata]
    key_prefix: str
    partition_name: str
    dataset_node_partition: DynamicPartitionsDefinition

    def __init__(self):
        self.key_prefix = self.country_metadata.id
        self.partition_name = f"{self.key_prefix}_nodes"
        self.dataset_node_partition = DynamicPartitionsDefinition(
            name=self.partition_name
        )

    def add_partition_keys(self, context, keys: list[str]):
        context.instance.add_dynamic_partitions(
            partitions_def_name=self.partition_name,
            partition_keys=keys,
        )

    def remove_all_partition_keys(self, context):
        for partition_key in context.instance.get_dynamic_partitions(
            self.partition_name
        ):
            context.instance.delete_dynamic_partition(
                self.partition_name, partition_key
            )

    def create_catalog(self):
        """Creates an asset providing a census metedata catalog."""

        @asset(key_prefix=self.key_prefix)
        def catalog(context) -> pd.DataFrame:
            return self._catalog(context)

        return catalog

    @abstractmethod
    def _catalog(self, context) -> pd.DataFrame:
        ...

    def create_country_metadata(self):
        """Creates an asset providing the country metadata."""

        @send_to_metadata_sensor
        @asset(key_prefix=self.key_prefix)
        def country_metadata(context):
            return self._country_metadata(context)

        return country_metadata

    def _country_metadata(self, _context) -> CountryMetadata:
        return self.country_metadata

    def create_data_publisher(self):
        """Creates an asset providing the data publisher metadata."""

        @send_to_metadata_sensor
        @asset(key_prefix=self.key_prefix)
        def data_publisher(context, country_metadata: CountryMetadata):
            return self._data_publisher(context, country_metadata)

        return data_publisher

    @abstractmethod
    def _data_publisher(
        self, context, country_metdata: CountryMetadata
    ) -> DataPublisher:
        ...

    def create_geometry(self):
        """
        Creates an asset providing a list of geometries, metadata and names
        at different resolutions.
        """

        @send_to_geometry_sensor
        @asset(key_prefix=self.key_prefix)
        def geometry(context):
            return self._geometry(context)

        return geometry

    @abstractmethod
    def _geometry(self, context) -> list[GeometryOutput]:
        ...

    def create_source_data_releases(self):
        """
        Creates an asset providing the corresponding source data release metadata for
        each geometry.
        """

        @send_to_metadata_sensor
        @asset(key_prefix=self.key_prefix)
        def source_data_releases(
            context,
            geometry: list[GeometryOutput],
            data_publisher: DataPublisher,
        ) -> dict[str, SourceDataRelease]:
            return self._source_data_releases(context, geometry, data_publisher)

        return source_data_releases

    @abstractmethod
    def _source_data_releases(
        self,
        context,
        geometry: list[GeometryOutput],
        data_publisher: DataPublisher,
        # TODO: consider version without inputs so only output type specified
        # **kwargs,
    ) -> dict[str, SourceDataRelease]:
        ...

    def create_census_tables(self):
        """
        Creates an asset providing each census table as a dataframe for each
        partition.
        """

        @asset(partitions_def=self.dataset_node_partition, key_prefix=self.key_prefix)
        def census_tables(context, catalog: pd.DataFrame) -> pd.DataFrame:
            return self._census_tables(context, catalog)

        return census_tables

    @abstractmethod
    def _census_tables(self, context, catalog: pd.DataFrame) -> pd.DataFrame:
        ...

    def create_source_metric_metadata(self):
        """
        Creates an asset providing the metadata required for downstream metric
        derivation.
        """

        @asset(partitions_def=self.dataset_node_partition, key_prefix=self.key_prefix)
        def source_metric_metadata(
            context, catalog, source_data_releases: dict[str, SourceDataRelease]
        ) -> MetricMetadata:
            return self._source_metric_metadata(context, catalog, source_data_releases)

        return source_metric_metadata

    @abstractmethod
    def _source_metric_metadata(
        self,
        context,
        catalog: pd.DataFrame,
        source_data_releases: dict[str, SourceDataRelease],
    ) -> MetricMetadata:
        ...

    def create_derived_metrics(self):
        """
        Creates an asset providing the metrics derived from the census tables and the
        corresponding source metric metadata.
        """

        @asset(partitions_def=self.dataset_node_partition, key_prefix=self.key_prefix)
        def derived_metrics(
            context,
            census_tables: pd.DataFrame,
            source_metric_metadata: MetricMetadata,
        ) -> MetricsOutput:
            return self._derived_metrics(context, census_tables, source_metric_metadata)

        return derived_metrics

    @abstractmethod
    def _derived_metrics(
        self,
        context,
        census_tables: pd.DataFrame,
        source_metric_metadata: MetricMetadata,
    ) -> MetricsOutput:
        ...

    def create_metrics(
        self,
        partitions_to_publish: list[str] | None = None,
    ):
        """
        Creates an asset combining all partitions across census tables into a combined
        list of metric data file names (for output), list of metadata and metric
        dataframe.
        """

        if partitions_to_publish is None:
            # Since this asset is unpartitioned, if the partition_mapping for
            # the upstream asset is not specified, Dagster assumes that this
            # asset depends on all upstream partitions.
            partition_kwargs = {}
        else:
            partition_kwargs = {
                "partition_mapping": SpecificPartitionsPartitionMapping(
                    partitions_to_publish
                )
            }

        @send_to_metrics_sensor
        @asset(
            key_prefix=self.key_prefix,
            ins={
                "derived_metrics": AssetIn(
                    key_prefix=self.key_prefix,
                    **partition_kwargs,  # pyright: ignore [reportArgumentType]
                )
            },
        )
        def metrics(
            context,
            # In principle, the derived_metrics should have the type:
            #    dict[str, MetricsOutput] | MetricsOutput
            # But Dagster doesn't like union types so we just use Any.
            derived_metrics,
        ) -> list[MetricsOutput]:
            # If the input asset has multiple partitions, Dagster returns a
            # dictionary of {partition_key: output_value}. However, if only one
            # partition is required, Dagster passes only the output value of
            # that partition, instead of a dictionary with one key. In this
            # case, we need to reconstruct a dictionary to pass it to the
            # underlying method which expects a dictionary.
            # See: https://github.com/dagster-io/dagster/issues/15538
            if partitions_to_publish is None:
                partition_names = context.instance.get_dynamic_partitions(
                    self.partition_name
                )
            else:
                partition_names = partitions_to_publish
            if len(partition_names) == 1:
                derived_metrics = {partition_names[0]: derived_metrics}

            return self._metrics(context, derived_metrics)

        return metrics

    def _metrics(
        self,
        context,
        derived_metrics: dict[str, MetricsOutput],
    ) -> list[MetricsOutput]:
        """
        Method which aggregates all the outputs of the `derived_metrics` asset.
        The default implementation simply combines each partition's output into
        a list (ignoring any partitions which have empty outputs). This is
        typically all that is required; however, this method can be overridden
        if different logic is required.
        """
        outputs = [
            output for output in derived_metrics.values() if len(output.metadata) > 0
        ]
        context.add_output_metadata(
            metadata={
                "num_metrics": sum(len(output.metadata) for output in outputs),
                "num_parquets": len(outputs),
            },
        )
        return outputs



from __future__ import annotations

from pathlib import Path

import fiona
import geopandas as gpd
import pandas as pd
import requests
from dagster import (
    AssetIn,
    AssetOut,
    MarkdownMetadataValue,
    MetadataValue,
    Output,
    asset,
    multi_asset,
)
from slugify import slugify

from poppusher.utils import (
    SourceDataAssumptionsOutdated,
    StagingDirResource,
    extract_main_file_from_zip,
    get_staging_dir,
    markdown_from_plot,
)

all_datasets = [
    {
        "Description": "1:250 000 Scale Colour Raster",
        "download_url": "https://api.os.uk/downloads/v1/products/250kScaleColourRaster/downloads?area=GB&format=TIFF-LZW&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/250kScaleColourRaster",
    },
    {
        "Description": "Boundary-Line",
        "download_url": "https://api.os.uk/downloads/v1/products/BoundaryLine/downloads?area=GB&format=GeoPackage&redirect",
        # "https://api.os.uk/downloads/v1/products/BoundaryLine/downloads?area=GB&format=Vector+Tiles&subformat=%28MBTiles%29&redirect"
        "landpage_url": "https://osdatahub.os.uk/downloads/open/BoundaryLine",
    },
    {
        "Description": "Code-Point® Open",
        "download_url": "https://api.os.uk/downloads/v1/products/CodePointOpen/downloads?area=GB&format=GeoPackage&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/CodePointOpen",
    },
    {
        "Description": "GB Overview Maps",
        "download_url": "https://api.os.uk/downloads/v1/products/GBOverviewMaps/downloads?area=GB&format=GeoTIFF&subformat=Full+Colour&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/GBOverviewMaps",
    },
    {
        "Description": "MiniScale®",
        "download_url": "https://api.os.uk/downloads/v1/products/MiniScale/downloads?area=GB&format=Zip+file+%28containing+EPS%2C+Illustrator+and+TIFF-LZW%29&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/MiniScale",
    },
    {
        "Description": "OS Open Built Up Areas",
        "download_url": "https://api.os.uk/downloads/v1/products/BuiltUpAreas/downloads?area=GB&format=GeoPackage&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/BuiltUpAreas",
    },
    {
        "Description": "OS Open Greenspace",
        "download_url": "https://api.os.uk/downloads/v1/products/OpenGreenspace/downloads?area=GB&format=GeoPackage&redirect",
        # "https://api.os.uk/downloads/v1/products/OpenGreenspace/downloads?area=GB&format=Vector+Tiles&subformat=%28MBTiles%29&redirect"
        "landpage_url": "https://osdatahub.os.uk/downloads/open/OpenGreenspace",
    },
    {
        "Description": "OS Open Linked Identifiers",
        # "Many download_urlS!"
        "download_url": None,
        "landpage_url": "https://osdatahub.os.uk/downloads/open/LIDS",
    },
    {
        "Description": "OS Open Names",
        "download_url": "https://api.os.uk/downloads/v1/products/OpenNames/downloads?area=GB&format=GeoPackage&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/OpenNames",
    },
    {
        "Description": "OS Open Rivers",
        "download_url": "https://api.os.uk/downloads/v1/products/OpenRivers/downloads?area=GB&format=GeoPackage&redirect",
        # "https://api.os.uk/downloads/v1/products/OpenRivers/downloads?area=GB&format=Vector+Tiles&subformat=%28MBTiles%29&redirect"
        "landpage_url": "https://osdatahub.os.uk/downloads/open/OpenRivers",
    },
    {
        "Description": "OS Open Roads",
        "download_url": "https://api.os.uk/downloads/v1/products/OpenRoads/downloads?area=GB&format=GeoPackage&redirect",
        # "https://api.os.uk/downloads/v1/products/OpenRoads/downloads?area=GB&format=Vector+Tiles&subformat=%28MBTiles%29&redirect"
        "landpage_url": "https://osdatahub.os.uk/downloads/open/OpenRoads",
    },
    {
        "Description": "OS Open TOID",
        # "Download per OSGB Grid Tile {eg 100km x 100km}"
        "download_url": None,
        "landpage_url": "https://osdatahub.os.uk/downloads/open/OpenTOID",
    },
    {
        "Description": "OS Open UPRN",
        "download_url": "https://api.os.uk/downloads/v1/products/OpenUPRN/downloads?area=GB&format=GeoPackage&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/OpenUPRN",
    },
    {
        "Description": "OS Open USRN",
        "download_url": "https://api.os.uk/downloads/v1/products/OpenUSRN/downloads?area=GB&format=GeoPackage&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/OpenUSRN",
    },
    {
        "Description": "OS Open Zoomstack",
        "download_url": "https://api.os.uk/downloads/v1/products/OpenZoomstack/downloads?area=GB&format=GeoPackage&redirect",
        # "https://api.os.uk/downloads/v1/products/OpenZoomstack/downloads?area=GB&format=Vector+Tiles&subformat=%28MBTiles%29&redirect"
        "landpage_url": "https://osdatahub.os.uk/downloads/open/OpenZoomstack",
    },
    {
        "Description": "OS OpenMap - Local",
        "download_url": "https://api.os.uk/downloads/v1/products/OpenMapLocal/downloads?area=GB&format=GeoPackage&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/OpenMapLocal",
    },
    {
        "Description": "OS Terrain® 50",
        "download_url": "https://api.os.uk/downloads/v1/products/Terrain50/downloads?area=GB&format=GeoPackage&redirect",
        # "https://api.os.uk/downloads/v1/products/Terrain50/downloads?area=GB&format=Vector+Tiles&subformat=%28MBTiles%29&redirect"
        "landpage_url": "https://osdatahub.os.uk/downloads/open/Terrain50",
    },
    {
        "Description": "OS VectorMap® District",
        "download_url": "https://api.os.uk/downloads/v1/products/VectorMapDistrict/downloads?area=GB&format=GeoPackage&redirect",
        "landpage_url": "https://osdatahub.os.uk/downloads/open/VectorMapDistrict",
    },
]

_all_datasets = []
for ds_dict in all_datasets:
    ds_dict["name"] = slugify(ds_dict["Description"], separator="_")
    _all_datasets.append(ds_dict)

all_datasets = _all_datasets

dataset_names = [dataset["name"] for dataset in all_datasets]


boundary_line_layers = [
    "boundary_line_ceremonial_counties",
    "boundary_line_historic_counties",
    "community_ward",
    "country_region",
    "county",
    "county_electoral_division",
    "district_borough_unitary",
    "district_borough_unitary_ward",
    "english_region",
    "greater_london_const",
    "high_water",
    "historic_european_region",
    "parish",
    "polling_districts_england",
    "scotland_and_wales_const",
    "scotland_and_wales_region",
    "unitary_electoral_division",
    "westminster_const",
]


@asset(description="List of all OS OpenData datasets")
def os_opendata_list():
    # Functionally this is just a passthrough, but it's useful to have an asset which
    # provides a "root" for all of the OS OpenData datasets.
    return pd.DataFrame(all_datasets)


@multi_asset(
    outs={name: AssetOut(key=name, is_required=False) for name in dataset_names},
    can_subset=True,
)
def uk_os_opendata(context, os_opendata_list):
    for ds_name in context.selected_output_names:
        ds_details = os_opendata_list[os_opendata_list["name"] == ds_name].to_dict(
            orient="records"
        )[0]
        yield Output(value=ds_details, output_name=ds_name)


@multi_asset(
    ins={"boundary_line": AssetIn("boundary_line")},
    outs={name: AssetOut(is_required=False) for name in boundary_line_layers},
    can_subset=True,
    required_resource_keys={"staging_res"},
)
def uk_os_opendata_boundary_line(context, boundary_line):
    # "Boundary-Line"
    # row = uk_os_opendata_list[uk_os_opendata_list["name"] == "Boundary-Line"]
    # download_url = row["download_url"].values[0]
    download_url = boundary_line["download_url"]
    staging_res = context.resources.staging_res

    archive_path_guess = "Data/file.gpkg"
    gpkg_path = download_with_redirect(
        context, staging_res, download_url, archive_path_guess
    )

    # TODO: Replace with:
    # https://pyogrio.readthedocs.io/en/latest/introduction.html#list-available-layers
    all_layers = set(fiona.listlayers(gpkg_path))

    expected_layers = set(boundary_line_layers)

    # Verify that the downloaded GeoPackage contains exactly the expected layers
    if all_layers != expected_layers:
        missing_layers = expected_layers - all_layers
        err_msg = ""
        if len(missing_layers) > 0:
            err_msg += (
                "Downloaded Boundary-Line GeoPackage is missing some expected layers:\n"
            )
            err_msg += "".join([f"- {layer}\n" for layer in missing_layers])

        extra_layers = all_layers - expected_layers

        if len(extra_layers) > 0:
            err_msg += (
                "Downloaded Boundary-Line GeoPackage has some unexpected layers:\n"
            )
            err_msg += "".join([f"- {layer}\n" for layer in extra_layers])

        err_msg += "Please check the download and update this DAG accordingly."
        raise SourceDataAssumptionsOutdated(err_msg)

    for lyr_name in context.selected_output_names:
        gdf = get_layer_from_gpkg(context, gpkg_path, lyr_name)
        yield Output(value=gdf, output_name=lyr_name)


def get_layer_from_gpkg(context, gpkg_path, layer_name):
    """
    Helper function to read a layer from a GeoPackage using pyogrio.
    """
    lyr_gdf = gpd.read_file(
        gpkg_path,
        layer=layer_name,
        # engine="pyogrio"
    )

    metadata = {
        "title": layer_name,
        "num_records": len(lyr_gdf),  # Metadata can be any key-value pair
        "columns": MetadataValue.md(
            "\n".join([f"- '`{col}`'" for col in lyr_gdf.columns.to_list()])
        ),
        "preview": MetadataValue.md(
            lyr_gdf.loc[:, lyr_gdf.columns != "geometry"].head().to_markdown()
        ),
    }

    # Only plot the layer if it's small enough
    if len(lyr_gdf) < 10000:
        # Plot and convert the image to Markdown to preview it within Dagster
        ax = lyr_gdf.plot(legend=False)
        ax.set_title(layer_name)
        md_plot = markdown_from_plot()
        # Force the type to be MarkdownMetadataValue
        # else sometimes Dagster will try to interpret it as a tuple for some unknown reason
        mdv: MarkdownMetadataValue = MetadataValue.md(md_plot)
        metadata["plot"] = mdv
    else:
        metadata["plot"] = MetadataValue.md(
            f"Too many features in the layer '{layer_name}' to plot"
        )

    context.add_output_metadata(metadata=metadata, output_name=layer_name)

    return lyr_gdf


def download_with_redirect(
    context,
    staging_res: StagingDirResource,
    source_download_url,
    source_archive_file_path,
):
    expected_extension = Path(source_archive_file_path).suffix

    with get_staging_dir(context, staging_res) as staging_dir_str:
        staging_path = Path(staging_dir_str)
        temp_file = staging_path / "temp_file.zip"

        # Given we're using a staging directory, we can assume that we only need to download the file if it doesn't exist
        if not temp_file.exists():
            # Using a custom header here to avoid a javascript redirect
            # https://stackoverflow.com/a/50111203
            headers = {"User-Agent": "poppusher"}

            with temp_file.open(mode="wb") as f, requests.get(
                source_download_url, headers=headers, allow_redirects=True, stream=True
            ) as r:
                for chunk in r.iter_content(chunk_size=(16 * 1024 * 1024)):
                    f.write(chunk)

        # unzip the file
        return extract_main_file_from_zip(temp_file, staging_path, expected_extension)



from __future__ import annotations

import importlib.metadata

import pytest

import poppusher as m


def test_version():
    assert importlib.metadata.version("poppusher") == m.__version__


def test_always_passing():
    assert True


@pytest.mark.xfail()
def test_always_failing():
    pytest.fail(
        "A test which always fails to check the CI infrastructure is working as expected"
    )



from __future__ import annotations

from poppusher.metadata import CountryMetadata

country: CountryMetadata = CountryMetadata(
    name_short_en="United Kingdom",
    name_official="The United Kingdom of Great Britain and Northern Ireland",
    iso3="GBR",
    iso2="GB",
    iso3166_2=None,
)

asset_prefix = "uk"



from __future__ import annotations

from pathlib import Path

from dagster import build_op_context
from python.poppusher.utils import StagingDirResource

from poppusher.utils import get_staging_dir


# Test StagingDirectory
def test_staging_directory(tmp_path):
    # Case A
    # Test that if stage_dir is found, then it is used
    context = build_op_context(
        # resources={"staging_dir": str(tmp_path / "mock_staging_dir")},
        partition_key="a/b",
        op_config={"unit_test_key": "test_staging_directory"},
    )

    staging_res = StagingDirResource(staging_dir=str(tmp_path / "mock_staging_dir"))

    with get_staging_dir(context, staging_res) as staging_dir_str:
        staging_dir = Path(staging_dir_str)
        assert staging_dir.exists()

        assert (
            staging_dir
            == tmp_path / "mock_staging_dir" / "test_staging_directory" / "a" / "b"
        )
        assert tmp_path in staging_dir.parents

    # Case B
    # Test that if stage_dir not found, then an unrelated temporary directory is used
    context = build_op_context(
        # resources={"staging_dir": None},
        partition_key="a/b",
    )

    staging_res = StagingDirResource(staging_dir=None)

    with get_staging_dir(context, staging_res) as staging_dir_str:
        staging_dir = Path(staging_dir_str)
        assert staging_dir.exists()

        assert "mock_staging_dir" not in staging_dir_str
        assert staging_dir != tmp_path
        assert tmp_path not in staging_dir.parents



from __future__ import annotations

import os
from collections.abc import Callable
from dataclasses import dataclass
from datetime import date

import pandas as pd
from dagster import AssetIn, MetadataValue, SpecificPartitionsPartitionMapping, asset
from icecream import ic
from rdflib import Graph, URIRef
from rdflib.namespace import DCAT, DCTERMS

from poppusher.assets.bel.utils import (
    DOWNLOAD_HANDLERS,
    check_not_str,
    check_str,
    filter_by_language,
    get_distribution_url,
    get_landpage_url,
    married_status_to_string,
    nationality_to_string,
    no_op_format_handler,
)
from poppusher.assets.country import Country
from poppusher.cloud_outputs import (
    GeometryOutput,
    MetricsOutput,
    send_to_geometry_sensor,
)
from poppusher.metadata import (
    COL,
    CountryMetadata,
    DataPublisher,
    GeometryMetadata,
    MetricMetadata,
    SourceDataRelease,
    metadata_to_dataframe,
)
from poppusher.utils import add_metadata, markdown_from_plot

KNOWN_FAILING_DATASETS = {
    # sqlite compressed as tar.gz
    "595",  # Census 2011 - Matrix of commutes by statistical sector
    # faulty zip file (confirmed by manual download)
    "2676",
    # Excel only (French and Dutch only)
    "2654",  # Geografische indelingen 2020
    "3961",  # Geografische indelingen 2021
    # AccessDB only!
    "4135",  # Enterprises subject to VAT according to legal form (English only)
    "4136",  # Enterprises subject to VAT according to employer class (English only)
}


@dataclass
class DatasetSpecification:
    # The geography level that the dataset is at. Must be one of the keys of
    # `BELGIUM_GEOMETRY_LEVELS`.
    geography_level: str
    # The column in the source data that contains the metric of interest
    source_column: str
    # The column in the source data that contains the geoID
    geoid_column: str
    # The columns that need to be pivoted to generate the derived metrics. If
    # this is empty, no pivoting is done, and the only metrics that are
    # published will be that of the `source_column`.
    pivot_columns: list[str]
    # The HXL tag for the derived metrics. If `pivot_columns` is empty, this
    # should just be a single string, which is the HXL tag of the original
    # `source_column`. If `pivot_columns` is not empty, this should be a
    # function which takes the values of the pivot columns in the correct
    # order. See below for an example.
    derived_hxl: str | Callable
    # The human readable name(s) for the derived metrics. Follows the same
    # rules as `derived_hxl`.
    derived_human_readable_name: str | Callable


DATASET_SPECIFICATIONS = {
    # Not actually census table -- these are the geometries from 2023
    "4726": DatasetSpecification(
        geography_level="statistical_sector",  # Lowest level available
        source_column="",
        geoid_column="",
        pivot_columns=[],
        derived_hxl="",
        derived_human_readable_name="",
    ),
    # Population in statistical sectors, 2023
    "4796": DatasetSpecification(
        geography_level="statistical_sector",
        source_column="TOTAL",
        geoid_column="CD_SECTOR",
        pivot_columns=[],
        derived_hxl="#population+adm5+total+2023",
        derived_human_readable_name="Population, total, 2023",
    ),
    # Population by nationality, marital status, age, and sex in
    # municipalities, 2023
    "4689": DatasetSpecification(
        geography_level="municipality",
        source_column="MS_POPULATION",
        geoid_column="CD_REFNIS",
        pivot_columns=["CD_NATLTY", "CD_CIV_STS", "CD_AGE", "CD_SEX"],
        derived_hxl=lambda n, ms, a, s: f"#population+adm4+{s.lower()}+age{a}+{nationality_to_string(n).lower()}+{married_status_to_string(ms).lower()}",
        derived_human_readable_name=lambda n, ms, a, s: f"Population, {nationality_to_string(n)}, {married_status_to_string(ms)}, {'female' if s == 'F' else 'male'}, and age {a}",
    ),
}

REQUIRED_DATASETS = (
    None if os.getenv("ENV") == "PROD" else list(DATASET_SPECIFICATIONS.keys())
)


@dataclass
class BelgiumGeometryLevel:
    level: str
    hxl_tag: str
    geo_id_column: str
    name_columns: dict[str, str]  # keys = language codes, values = column names


BELGIUM_GEOMETRY_LEVELS = {
    "province": BelgiumGeometryLevel(
        level="province",
        hxl_tag="adm1",
        geo_id_column="cd_prov_refnis",
        name_columns={
            "nld": "tx_prov_descr_nl",
            "fra": "tx_prov_descr_fr",
            "deu": "tx_prov_descr_de",
        },
    ),
    "region": BelgiumGeometryLevel(
        level="region",
        hxl_tag="adm2",
        geo_id_column="cd_rgn_refnis",
        name_columns={
            "nld": "tx_rgn_descr_nl",
            "fra": "tx_rgn_descr_fr",
            "deu": "tx_rgn_descr_de",
        },
    ),
    "arrondisement": BelgiumGeometryLevel(
        level="arrondisement",
        hxl_tag="adm3",
        geo_id_column="cd_dstr_refnis",
        name_columns={
            "nld": "tx_adm_dstr_descr_nl",
            "fra": "tx_adm_dstr_descr_fr",
            "deu": "tx_adm_dstr_descr_de",
        },
    ),
    "municipality": BelgiumGeometryLevel(
        level="municipality",
        hxl_tag="adm4",
        geo_id_column="cd_munty_refnis",
        name_columns={
            "nld": "tx_munty_descr_nl",
            "fra": "tx_munty_descr_fr",
            "deu": "tx_munty_descr_de",
        },
    ),
    "statistical_sector": BelgiumGeometryLevel(
        level="statistical_sector",
        hxl_tag="adm5",
        geo_id_column="cd_sector",
        name_columns={
            "nld": "tx_sector_descr_nl",
            "fra": "tx_sector_descr_fr",
            "deu": "tx_sector_descr_de",
        },
    ),
}


class Belgium(Country):
    country_metadata: CountryMetadata = CountryMetadata(
        name_short_en="Belgium",
        name_official="Kingdom of Belgium",
        iso3="BEL",
        iso2="BE",
        iso3166_2=None,
    )

    def _country_metadata(self, _context) -> CountryMetadata:
        return self.country_metadata

    def _data_publisher(
        self, _context, country_metadata: CountryMetadata
    ) -> DataPublisher:
        return DataPublisher(
            name="Statbel",
            url="https://statbel.fgov.be/en",
            description="Statbel is the Belgian statistical office. It is part of the Federal Public Service Economy, SMEs, Self-employed and Energy.",
            countries_of_interest=[country_metadata.id],
        )

    @staticmethod
    def get_opendata_dataset_list() -> Graph:
        """
        Returns a list of all the tables available in the Statbel Open Data portal.

        This document is essential reading for understanding the structure of the data:
        https://github.com/belgif/inspire-dcat/blob/main/DCATAPprofil.en.md
        """
        CATALOG_URL = (
            "https://doc.statbel.be/publications/DCAT/DCAT_opendata_datasets.ttl"
        )
        graph = Graph()
        graph.parse(CATALOG_URL, format="ttl")
        return graph

    def _catalog(self, context) -> pd.DataFrame:
        self.remove_all_partition_keys(context)

        opendata_dataset_list = self.get_opendata_dataset_list()
        # Create the schema for the catalog
        catalog_summary = {
            "node": [],
            "human_readable_name": [],
            "description": [],
            "metric_parquet_path": [],
            "parquet_column_name": [],
            "parquet_margin_of_error_column": [],
            "parquet_margin_of_error_file": [],
            "potential_denominator_ids": [],
            "parent_metric_id": [],
            "source_data_release_id": [],
            "source_download_url": [],
            "source_format": [],
            "source_archive_file_path": [],
            "source_documentation_url": [],
        }

        # Loop over the datasets in the catalogue Graph
        CATALOG_ROOT = URIRef("http://data.gov.be/catalog/statbelopen")
        for dataset_id in opendata_dataset_list.objects(
            subject=CATALOG_ROOT, predicate=DCAT.dataset, unique=True
        ):
            catalog_summary["node"].append(
                str(dataset_id).removeprefix("https://statbel.fgov.be/node/")
            )
            catalog_summary["human_readable_name"].append(
                filter_by_language(
                    graph=opendata_dataset_list,
                    subject=dataset_id,
                    predicate=DCTERMS.title,
                )
            )
            catalog_summary["description"].append(
                filter_by_language(
                    opendata_dataset_list,
                    subject=dataset_id,
                    predicate=DCTERMS.description,
                )
            )

            # This is unknown at this stage
            catalog_summary["metric_parquet_path"].append(None)
            catalog_summary["parquet_margin_of_error_column"].append(None)
            catalog_summary["parquet_margin_of_error_file"].append(None)
            catalog_summary["potential_denominator_ids"].append(None)
            catalog_summary["parent_metric_id"].append(None)
            catalog_summary["source_data_release_id"].append(None)
            catalog_summary["parquet_column_name"].append(None)

            download_url, archive_file_path, format = get_distribution_url(
                opendata_dataset_list, dataset_id
            )
            catalog_summary["source_download_url"].append(download_url)
            catalog_summary["source_archive_file_path"].append(archive_file_path)
            catalog_summary["source_format"].append(format)

            catalog_summary["source_documentation_url"].append(
                get_landpage_url(opendata_dataset_list, dataset_id, language="en")
            )

        # Convert to dataframe and remove datasets which are known to fail
        catalog_df = pd.DataFrame(data=catalog_summary, dtype="string")
        catalog_df = catalog_df[~catalog_df["node"].isin(KNOWN_FAILING_DATASETS)]

        # If not production, restrict the catalog to only the required datasets
        if REQUIRED_DATASETS is not None:
            catalog_df = catalog_df[catalog_df["node"].isin(REQUIRED_DATASETS)]

        self.add_partition_keys(context, catalog_df["node"].to_list())
        add_metadata(context, catalog_df, "Catalog")
        return catalog_df

    def _census_tables(self, context, catalog: pd.DataFrame) -> pd.DataFrame:
        row = catalog[catalog["node"] == context.partition_key]
        format = row["source_format"].iloc[0]
        ic(format)
        handler = DOWNLOAD_HANDLERS.get(format, no_op_format_handler)
        return handler(context, row=row)

    def _geometry(self, _context) -> list[GeometryOutput]:
        err = "The geometry asset for Belgium has a custom implementation as it depends on the census_tables asset."
        raise NotImplementedError(err)

    def _source_data_releases(
        self, _context, geometry, data_publisher
    ) -> dict[str, SourceDataRelease]:
        return {
            geo_output.metadata.level: SourceDataRelease(
                name="StatBel Open Data",
                date_published=date(2015, 10, 22),
                reference_period_start=date(2015, 10, 22),
                reference_period_end=date(2015, 10, 22),
                collection_period_start=date(2015, 10, 22),
                collection_period_end=date(2015, 10, 22),
                expect_next_update=date(2022, 1, 1),
                url="https://statbel.fgov.be/en/open-data",
                description="TBC",
                data_publisher_id=data_publisher.id,
                geometry_metadata_id=geo_output.metadata.id,
            )
            for geo_output in geometry
        }

    def _source_metric_metadata(
        self,
        context,
        catalog: pd.DataFrame,
        source_data_releases: dict[str, SourceDataRelease],
    ) -> MetricMetadata:
        catalog_row = catalog[catalog["node"] == context.partition_key].to_dict(
            orient="records"
        )[0]
        dataset_spec = DATASET_SPECIFICATIONS[context.partition_key]

        return MetricMetadata(
            human_readable_name=catalog_row["human_readable_name"],
            source_download_url=catalog_row["source_download_url"],
            source_archive_file_path=catalog_row["source_archive_file_path"],
            source_documentation_url=catalog_row["source_documentation_url"],
            source_data_release_id=source_data_releases[
                dataset_spec.geography_level
            ].id,
            parent_metric_id=None,
            potential_denominator_ids=None,
            description=catalog_row["description"].strip(),
            source_metric_id=dataset_spec.source_column,
            # These are to be replaced at the derived stage
            metric_parquet_path="__PLACEHOLDER__",
            hxl_tag="__PLACEHOLDER__",
            parquet_column_name="__PLACEHOLDER__",
            parquet_margin_of_error_file=None,
            parquet_margin_of_error_column=None,
        )

    def _derived_metrics(
        self,
        context,
        census_tables: pd.DataFrame,
        source_metric_metadata: MetricMetadata,
    ) -> MetricsOutput:
        # Skip if we don't know what to do with this partition
        try:
            this_dataset_spec = DATASET_SPECIFICATIONS[context.partition_key]
        except KeyError:
            skip_reason = (
                f"No action specified for partition key {context.partition_key}"
            )
            context.log.warning(skip_reason)
            return MetricsOutput(metadata=[], metrics=pd.DataFrame())

        # Return empty metrics if the source table is empty, if the source
        # column is not present
        skip_reason = None
        if len(census_tables) == 0:
            skip_reason = "Skipping as input table is empty"
            context.log.warning(skip_reason)
            return MetricsOutput(metadata=[], metrics=pd.DataFrame())
        if this_dataset_spec.source_column not in census_tables.columns:
            skip_reason = (
                f"Skipping as source column '{this_dataset_spec.source_column}' is not"
                f" present in the input table."
            )
            context.log.warning(skip_reason)
            return MetricsOutput(metadata=[], metrics=pd.DataFrame())

        # Assign parquet file name
        parquet_file_name = (
            f"{self.country_metadata.id}/metrics/"
            f"{''.join(c for c in context.partition_key if c.isalnum()) + '.parquet'}"
        )

        # Rename the geoID column to GEO_ID
        geo_id_col_name = this_dataset_spec.geoid_column
        census_tables = census_tables.rename(
            columns={geo_id_col_name: COL.GEO_ID.value}
        )

        # Generate derived metrics through pivoting
        if len(this_dataset_spec.pivot_columns) > 0:
            needed_columns = [
                COL.GEO_ID.value,
                this_dataset_spec.source_column,
                *this_dataset_spec.pivot_columns,
            ]
            census_table = census_tables[needed_columns]
            census_table = census_table.pivot_table(
                index=COL.GEO_ID.value,
                columns=this_dataset_spec.pivot_columns,
                values=this_dataset_spec.source_column,
            )
            # Generate metadata structs
            derived_mmds = []
            for c in census_table.columns.to_flat_index():
                new_mmd = source_metric_metadata.copy()
                new_mmd.parent_metric_id = source_metric_metadata.source_metric_id
                new_mmd.metric_parquet_path = parquet_file_name
                check_not_str(this_dataset_spec.derived_hxl)
                new_mmd.hxl_tag = this_dataset_spec.derived_hxl(
                    *c
                )  # type: ignore[reportCallIssue]
                check_not_str(this_dataset_spec.derived_human_readable_name)
                new_mmd.human_readable_name = (
                    this_dataset_spec.derived_human_readable_name(*c)
                )  # type: ignore[reportCallIssue]
                new_mmd.parquet_column_name = "_".join([str(x) for x in c])
                derived_mmds.append(new_mmd)
            # Rename columns
            col_names = [m.parquet_column_name for m in derived_mmds]
            census_table.columns = col_names

        else:
            # No pivoting required. Just extract the column
            census_table = census_tables[
                [COL.GEO_ID.value, this_dataset_spec.source_column]
            ]
            # Generate metadata struct
            new_mmd = source_metric_metadata.copy()
            new_mmd.parent_metric_id = source_metric_metadata.source_metric_id
            new_mmd.metric_parquet_path = parquet_file_name
            check_str(this_dataset_spec.derived_hxl)
            new_mmd.hxl_tag = this_dataset_spec.derived_hxl  # type: ignore[reportAttributeAccessIssue]
            check_str(this_dataset_spec.derived_human_readable_name)
            new_mmd.human_readable_name = this_dataset_spec.derived_human_readable_name  # type: ignore[reportAttributeAccessIssue]
            new_mmd.parquet_column_name = this_dataset_spec.source_column
            derived_mmds = [new_mmd]

        context.add_output_metadata(
            metadata={
                "metadata_preview": MetadataValue.md(
                    metadata_to_dataframe(derived_mmds).head().to_markdown()
                ),
                "metrics_shape": f"{census_table.shape[0]} rows x {census_table.shape[1]} columns",
                "metrics_preview": MetadataValue.md(census_table.head().to_markdown()),
            },
        )
        return MetricsOutput(metadata=derived_mmds, metrics=census_table)


# Create assets
bel = Belgium()
country_metadata = bel.create_country_metadata()
data_publisher = bel.create_data_publisher()
catalog = bel.create_catalog()
census_tables = bel.create_census_tables()


@send_to_geometry_sensor
@asset(
    ins={
        "sector_geometries": AssetIn(
            key=[bel.country_metadata.id, "census_tables"],
            partition_mapping=SpecificPartitionsPartitionMapping(["4726"]),
        ),
    },
    key_prefix=bel.country_metadata.id,
)
def geometry(context, sector_geometries) -> list[GeometryOutput]:
    """
    Produces the full set of data / metadata associated with Belgian
    municipalities. The outputs, in order, are:

    1. A DataFrame containing a serialised GeometryMetadata object.
    2. A GeoDataFrame containing the geometries of the municipalities.
    3. A DataFrame containing the names of the municipalities (in this case,
       they are in Dutch, French, and German).
    """
    geometries_to_return = []

    for level_details in BELGIUM_GEOMETRY_LEVELS.values():
        geometry_metadata = GeometryMetadata(
            country_metadata=bel.country_metadata,
            validity_period_start=date(2023, 1, 1),
            validity_period_end=date(2023, 12, 31),
            level=level_details.level,
            hxl_tag=level_details.hxl_tag,
        )

        region_geometries = (
            sector_geometries.dissolve(by=level_details.geo_id_column)
            .reset_index()
            .rename(columns={level_details.geo_id_column: COL.GEO_ID.value})
            .loc[:, ["geometry", COL.GEO_ID.value]]
        )
        ic(region_geometries.head())

        region_names = (
            sector_geometries.rename(
                columns={
                    level_details.geo_id_column: COL.GEO_ID.value,
                    level_details.name_columns["nld"]: "nld",
                    level_details.name_columns["fra"]: "fra",
                    level_details.name_columns["deu"]: "deu",
                }
            )
            .loc[:, [COL.GEO_ID.value, "nld", "fra", "deu"]]
            .drop_duplicates()
            .astype({COL.GEO_ID.value: str})
        )
        ic(region_names.head())

        geometries_to_return.append(
            GeometryOutput(
                metadata=geometry_metadata, gdf=region_geometries, names_df=region_names
            )
        )

    # Add output metadata
    first_output = geometries_to_return[0]
    first_joined_gdf = first_output.gdf.merge(
        first_output.names_df, on=COL.GEO_ID.value
    )
    ax = first_joined_gdf.plot(column="nld", legend=False)
    ax.set_title(f"Belgium 2023 {first_output.metadata.level}")
    md_plot = markdown_from_plot()
    context.add_output_metadata(
        metadata={
            "all_geom_levels": MetadataValue.md(
                ",".join(
                    [geom_output.metadata.level for geom_output in geometries_to_return]
                )
            ),
            "first_geometry_plot": MetadataValue.md(md_plot),
            "first_names_preview": MetadataValue.md(
                first_output.names_df.head().to_markdown()
            ),
        }
    )

    return geometries_to_return


source_data_releases = bel.create_source_data_releases()
source_metric_metadata = bel.create_source_metric_metadata()
derived_metrics = bel.create_derived_metrics()
metrics = bel.create_metrics()



from __future__ import annotations

from pathlib import Path

import geopandas as gpd
import pytest
from rdflib import Graph


@pytest.fixture(scope="module")
def demo_sectors() -> gpd.GeoDataFrame:
    input_path = str(Path(__file__).parent / "demo_data" / "be_demo_sector.geojson")
    return gpd.read_file(input_path)


@pytest.fixture(scope="module")
def demo_catalog() -> Graph:
    input_path = str(
        Path(__file__).parent / "demo_data" / "statbel_opendata_subset.ttl"
    )

    graph = Graph()
    graph.parse(input_path, format="ttl")

    return graph


# # TODO: consider revising tests to incorporate changes following change to base class implementation.
# # See issue: https://github.com/Urban-Analytics-Technology-Platform/poppusher/issues/133
# @pytest.fixture(scope="module")
# def demo_catalog_df(demo_catalog) -> pd.DataFrame:
#     context = build_asset_context()
#     return bel.census_tables.catalog_as_dataframe(context, demo_catalog)


# @pytest.mark.skip(
#     reason="Need to re-implement aggregate_sectors_to_municipalities to work with the sectors coming from the partitioned asset."
# )
# def test_aggregate_sectors_to_municipalities(demo_sectors):
#     # Test the that the row count is correctly added to the metadata
#     context = build_asset_context()

#     actual_municipalities = bel.census_geometry.aggregate_sectors_to_municipalities(
#         context, demo_sectors
#     )

#     expected_sector_row_count = 7
#     expected_municipalities_row_count = 3

#     assert len(demo_sectors) == expected_sector_row_count
#     assert len(actual_municipalities) == expected_municipalities_row_count
#     metadata = context.get_output_metadata(output_name="result")
#     assert metadata["num_records"] == expected_municipalities_row_count


# @pytest.mark.skip(reason="Fix test_get_population_details_per_municipality first")
# def test_get_population_details_per_municipality():
#     with build_asset_context() as muni_context:
#         stat_muni = bel.census_tables.get_population_details_per_municipality(
#             muni_context
#         )

#     ic(len(stat_muni))
#     ic(stat_muni.columns)

#     assert len(stat_muni) > 0
#     assert len(stat_muni.columns) > 0

#     pytest.fail("Not complete")


# @pytest.mark.skip(reason="Fix test_get_population_details_per_municipality first")
# def test_pivot_population():
#     # Test the that the row count is correctly added to the metadata
#     # muni_context = build_asset_context()

#     with build_asset_context() as muni_context:
#         # Check that the metadata is empty initially
#         assert (muni_context.get_output_metadata(output_name="result") is None) | (
#             muni_context.get_output_metadata(output_name="result") == {}
#         )

#         # Get the geometries
#         stat_muni = bel.census_tables.get_population_details_per_municipality(
#             muni_context
#         )

#     assert len(stat_muni) > 0
#     ic(len(stat_muni))
#     ic(stat_muni.head())

#     # pivot_context = build_asset_context()

#     with build_asset_context() as pivot_context:
#         # Pivot the population
#         pivoted = bel.pivot_population(pivot_context, stat_muni)

#     expected_number_of_municipalities = 581

#     # Now check that the metadata has been updated
#     metadata = pivot_context.get_output_metadata(output_name="result")
#     assert len(pivoted) == expected_number_of_municipalities
#     assert metadata["num_records"] == expected_number_of_municipalities


# def test_demo_catalog(demo_catalog):
#     # There are 10 datasets in the demo catalogue
#     expected_length = 10
#     actual_length = len(
#         list(
#             demo_catalog.objects(
#                 subject=bel.census_tables.opendata_catalog_root,
#                 predicate=DCAT.dataset,
#                 unique=False,
#             )
#         )
#     )

#     assert actual_length == expected_length


# def test_catalog_metadata_details(demo_catalog_df):
#     # Get the metadata for a specific dataset in the demo catalogue:
#     # https://statbel.fgov.be/node/4151 "Population by Statistical sector"
#     # mmd = bel.census_tables.get_mmd_from_dataset_node(
#     #     demo_catalog, dataset_node=URIRef("https://statbel.fgov.be/node/4151")
#     # )

#     row = demo_catalog_df[
#         demo_catalog_df["node"].eq("https://statbel.fgov.be/node/4151")
#     ].to_dict(orient="records")[0]

#     # Check that the right distribution_url has been selected
#     #
#     # This dataset has two distributions:
#     # (xlsx):    <https://statbel.fgov.be/sites/default/files/files/opendata/bevolking/sectoren/OPENDATA_SECTOREN_2022.xlsx#distribution4151>,
#     # (txt/zip): <https://statbel.fgov.be/sites/default/files/files/opendata/bevolking/sectoren/OPENDATA_SECTOREN_2022.zip#distribution4151> ;
#     #
#     # We expect the txt/zip version to be selected.
#     expected_distribution_url = "https://statbel.fgov.be/sites/default/files/files/opendata/bevolking/sectoren/OPENDATA_SECTOREN_2022.zip#distribution4151"
#     wrong_distribution_url = "https://statbel.fgov.be/sites/default/files/files/opendata/bevolking/sectoren/OPENDATA_SECTOREN_2022.xlsx#distribution4151"

#     assert row["source_download_url"] == expected_distribution_url
#     assert row["source_download_url"] != wrong_distribution_url

#     # We expect the title to be in English (not any of the other available languages)
#     title_english = "Population by Statistical sector"
#     title_german = "Bevölkerung nach statistischen Sektoren"
#     title_french = "Population par secteur statistique"
#     title_dutch = "Bevolking per statistische sector"

#     assert row["human_readable_name"] == title_english
#     assert row["human_readable_name"] != title_german
#     assert row["human_readable_name"] != title_french
#     assert row["human_readable_name"] != title_dutch


# @pytest.mark.skip(reason="Test not implemented")
# def test_filter_by_language():
#     # Test case
#     # This dataset is only available in Dutch and French
#     # https://statbel.fgov.be/node/2654
#     pytest.fail("Not implemented")


# def test_catalog_as_dataframe(demo_catalog_df):
#     # Check that the catalog has been converted to a DataFrame
#     assert isinstance(demo_catalog_df, pd.DataFrame)

#     # Check that the DataFrame has the expected number of rows
#     expected_number_of_datasets = 10
#     assert len(demo_catalog_df) == expected_number_of_datasets

#     # # Convert the demo catalog to a DataFrame
#     # with build_asset_context() as context:
#     #     catalog_df = bel.census_tables.catalog_as_dataframe(context, demo_catalog_df)

#     #     # Check that the catalog has been converted to a DataFrame
#     #     assert isinstance(catalog_df, pd.DataFrame)

#     #     # Check that the DataFrame has the expected number of rows
#     #     expected_number_of_datasets = 10
#     #     assert len(catalog_df) == expected_number_of_datasets

#     #     # Also check that the metadata has been updated
#     #     metadata = context.get_output_metadata(output_name="result")
#     #     assert metadata["num_records"] == expected_number_of_datasets


# @pytest.mark.skip(reason="Test not implemented")
# def test_purepath_suffix():
#     # examples
#     # cases = [
#     #     (
#     #         "https://statbel.fgov.be/sites/default/files/files/opendata/bevolking/sectoren/OPENDATA_SECTOREN_2022.zip#distribution4151",
#     #         "zip",
#     #     ),
#     #     (
#     #         "https://statbel.fgov.be/sites/default/files/files/opendata/bevolking/sectoren/OPENDATA_SECTOREN_2022.xlsx#distribution4151",
#     #         "xlsx",
#     #     ),
#     #     (
#     #         "https://statbel.fgov.be/sites/default/files/files/opendata/bevolking/sectoren/OPENDATA_SECTOREN_2022.txt#distribution4151",
#     #         "txt",
#     #     ),
#     # ]
#     pytest.fail("Not implemented")


# def test_filter_known_failing_datasets():
#     mock_catalog = [
#         "https://statbel.fgov.be/node/4796",  # from census_derived._needed_dataset_nodes
#         "https://statbel.fgov.be/node/3961",  # Known failing dataset
#         "https://statbel.fgov.be/node/595",  # Known failing dataset
#         "https://statbel.fgov.be/en/node",  # Incomplete URL
#         "2676",  # Known failing dataset node number
#     ]

#     expected_list = [
#         "https://statbel.fgov.be/node/4796",
#         "https://statbel.fgov.be/en/node",
#         "2676",
#     ]

#     actual_list = bel.census_tables.filter_known_failing_datasets(mock_catalog)

#     assert mock_catalog != expected_list
#     assert actual_list != mock_catalog
#     assert actual_list == expected_list



from __future__ import annotations

from dataclasses import dataclass

import geopandas as gpd
import pandas as pd
from dagster import InputContext, IOManager, MetadataValue, OutputContext
from iso639 import iter_langs
from upath import UPath

from poppusher.cloud_outputs import GeometryOutput, MetricsOutput
from poppusher.metadata import (
    COL,
    CountryMetadata,
    DataPublisher,
    GeometryMetadata,
    MetricMetadata,
    SourceDataRelease,
    metadata_to_dataframe,
)

# Set of all valid ISO 639-3 codes. (They are all lowercase)
VALID_ISO639_3_CODES = {lang.pt3 for lang in iter_langs()}


class IOManagerError(Exception):
    pass


class PoppusherIOManager(IOManager):
    def get_base_path(self) -> UPath:
        raise NotImplementedError

    def handle_df(
        self, context: OutputContext, df: pd.DataFrame, full_path: UPath
    ) -> None:
        raise NotImplementedError

    def load_input(self, _context: InputContext) -> pd.DataFrame:
        err_msg = "This IOManager is only for writing outputs"
        raise RuntimeError(err_msg)


class MetadataIOManager(PoppusherIOManager):
    def get_output_filename(
        self, obj_entry: CountryMetadata | DataPublisher | SourceDataRelease
    ) -> str:
        if isinstance(obj_entry, CountryMetadata):
            return "country_metadata.parquet"
        if isinstance(obj_entry, DataPublisher):
            return "data_publishers.parquet"
        if isinstance(obj_entry, SourceDataRelease):
            return "source_data_releases.parquet"

        err_msg = "This IO manager only accepts CountryMetadata, DataPublisher, and SourceDataRelease"
        raise ValueError(err_msg)

    def get_full_path(
        self,
        context: OutputContext,
        obj_entry: CountryMetadata | DataPublisher | SourceDataRelease,
    ) -> UPath:
        path_prefixes = list(context.partition_key.split("/"))[:-1]
        filename = self.get_output_filename(obj_entry)
        return self.get_base_path() / UPath("/".join([*path_prefixes, filename]))

    def handle_output(
        self,
        context: OutputContext,
        obj: (
            CountryMetadata
            | DataPublisher
            | SourceDataRelease
            | list[CountryMetadata]
            | list[DataPublisher]
            | list[SourceDataRelease]
            | dict[str, CountryMetadata]
            | dict[str, DataPublisher]
            | dict[str, SourceDataRelease]
        ),
    ):
        def all_same_class_and_accepted(objs) -> bool:
            return len(objs) > 0 and (
                all(isinstance(obj, CountryMetadata) for obj in objs)
                or all(isinstance(obj, DataPublisher) for obj in objs)
                or all(isinstance(obj, SourceDataRelease) for obj in objs)
            )

        if isinstance(obj, dict) and all_same_class_and_accepted(obj.values()):
            vals = list(obj.values())
        elif isinstance(obj, list) and all_same_class_and_accepted(obj):
            vals = obj
        elif isinstance(obj, CountryMetadata | DataPublisher | SourceDataRelease):
            vals = [obj]
        else:
            err_msg = (
                "MetadataIOManager only accepts assets which produce"
                " individual `CountryMetadata`, `DataPublisher`, or"
                " `SourceDataRelease`; or lists of them; or dictionaries"
                " with them as the values. If lists or dictionaries are"
                " passed, they must be nonempty, and all items must be"
                " of the same type."
            )
            raise ValueError(err_msg)

        full_path = self.get_full_path(context, vals[0])
        context.add_output_metadata(metadata={"parquet_path": str(full_path)})
        self.handle_df(context, metadata_to_dataframe(vals), full_path)


class GeoIOManager(PoppusherIOManager):
    def handle_flatgeobuf(
        self, context: OutputContext, geo_df: gpd.GeoDataFrame, full_path: UPath
    ) -> None:
        raise NotImplementedError

    def handle_geojsonseq(
        self, context: OutputContext, geo_df: gpd.GeoDataFrame, full_path: UPath
    ) -> None:
        raise NotImplementedError

    def handle_pmtiles(
        self, _context: OutputContext, _geo_df: gpd.GeoDataFrame, _full_path: UPath
    ) -> None:
        err_msg = "Pmtiles not currently implemented. You shouldn't be calling this."
        raise RuntimeError(err_msg)

    @dataclass
    class GeometryOutputPaths:
        flatgeobuf: UPath
        pmtiles: UPath
        geojsonseq: UPath
        names: UPath

    def get_full_paths_geoms(
        self,
        geo_metadata: GeometryMetadata,
    ) -> GeometryOutputPaths:
        filepath_stem = geo_metadata.filepath_stem
        base_path = self.get_base_path()
        return self.GeometryOutputPaths(
            flatgeobuf=base_path / UPath(f"{filepath_stem}.fgb"),
            pmtiles=base_path / UPath(f"TODO_{filepath_stem}.pmtiles"),
            geojsonseq=base_path / UPath(f"{filepath_stem}.geojsonseq"),
            names=base_path / UPath(f"{filepath_stem}.parquet"),
        )

    def get_full_path_metadata(
        self,
        context: OutputContext,
    ) -> UPath:
        base_path = self.get_base_path()
        asset_prefix = list(context.partition_key.split("/"))[:-1]
        return base_path / UPath("/".join([*asset_prefix, "geometry_metadata.parquet"]))

    def handle_output(
        self,
        context: OutputContext,
        obj: list[GeometryOutput],
    ) -> None:
        output_metadata = {
            "flatgeobuf_paths": [],
            "pmtiles_paths": [],
            "geojsonseq_paths": [],
            "names_paths": [],
        }

        for output in obj:
            # Perform checks on input data
            if set(output.gdf.columns) != {COL.GEO_ID.value, "geometry"}:
                err_msg = (
                    f"The geodataframe passed to GeometryIOManager must"
                    f" have only two columns, '{COL.GEO_ID.value}' and"
                    f" 'geometry'."
                )
                raise ValueError(err_msg)
            names_cols = set(output.names_df.columns)
            if COL.GEO_ID.value not in names_cols:
                # Check if it's the index, reset index if so
                if output.names_df.index.name == COL.GEO_ID.value:
                    output.names_df = output.names_df.reset_index()
                    names_cols = set(output.names_df.columns)
                else:
                    err_msg = (
                        f"The dataframe of names passed to GeometryIOManager"
                        f" must have a '{COL.GEO_ID.value}' column."
                    )
                    raise ValueError(err_msg)
            other_names_cols = names_cols - {COL.GEO_ID.value}
            if len(other_names_cols) == 0:
                err_msg = (
                    f"The dataframe of names passed to GeometryIOManager"
                    f" must have at least one column other than"
                    f" '{COL.GEO_ID.value}'."
                )
                raise ValueError(err_msg)
            not_iso639_3_cols = [
                col for col in other_names_cols if col not in VALID_ISO639_3_CODES
            ]
            if len(not_iso639_3_cols) > 0:
                err_msg = (
                    f"The names of the column(s)"
                    f" {not_iso639_3_cols} in the dataframe of"
                    f" names passed to GeometryIOManager are not"
                    f" valid ISO 639-3 codes."
                )
                raise ValueError(err_msg)
            # Coerce columns to strings (pandas doesn't do this automatically)
            output.gdf[COL.GEO_ID.value] = output.gdf[COL.GEO_ID.value].astype("string")
            output.names_df = output.names_df.astype("string")

            full_paths = self.get_full_paths_geoms(output.metadata)

            self.handle_flatgeobuf(context, output.gdf, full_paths.flatgeobuf)
            self.handle_geojsonseq(context, output.gdf, full_paths.geojsonseq)
            # TODO self.handle_pmtiles(context, output.gdf, full_paths.pmtiles)
            self.handle_df(context, output.names_df, full_paths.names)

            output_metadata["flatgeobuf_paths"].append(str(full_paths.flatgeobuf))
            output_metadata["pmtiles_paths"].append(str(full_paths.pmtiles))
            output_metadata["geojsonseq_paths"].append(str(full_paths.geojsonseq))
            output_metadata["names_paths"].append(str(full_paths.names))

        metadata_df_filepath = self.get_full_path_metadata(context)
        metadata_df = metadata_to_dataframe([geo_output.metadata for geo_output in obj])
        self.handle_df(context, metadata_df, metadata_df_filepath)

        context.add_output_metadata(
            metadata={
                "metadata_path": str(metadata_df_filepath),
                "metadata_preview": MetadataValue.md(metadata_df.head().to_markdown()),
                **output_metadata,
            }
        )


class MetricsIOManager(PoppusherIOManager):
    def get_full_path_metadata(
        self,
        context: OutputContext,
    ) -> UPath:
        base_path = self.get_base_path()
        asset_prefix = list(context.partition_key.split("/"))[:-1]
        return base_path / UPath("/".join([*asset_prefix, "metric_metadata.parquet"]))

    def get_full_path_metrics(
        self,
        parquet_path: str,
    ) -> UPath:
        base_path = self.get_base_path()
        return base_path / UPath(parquet_path)

    def handle_output(
        self,
        context: OutputContext,
        obj: list[MetricsOutput],
    ) -> None:
        # Perform checks on input data
        all_filepaths = []
        for output in obj:
            # Check GEO_ID col
            metrics_cols = set(output.metrics.columns)
            if COL.GEO_ID.value not in metrics_cols:
                # Check if it's the index, reset index if so
                if output.metrics.index.name == COL.GEO_ID.value:
                    output.metrics = output.metrics.reset_index()
                    metrics_cols = set(output.metrics.columns)
                else:
                    err_msg = (
                        f"The dataframe of metrics passed to MetricsIOManager"
                        f" must have a '{COL.GEO_ID.value}' column. It only has"
                        f" columns {metrics_cols}."
                    )
                    raise ValueError(err_msg)
            other_metrics_cols = metrics_cols - {COL.GEO_ID.value}
            # The data column names must match the metadata
            metadata_cols = {mmd.parquet_column_name for mmd in output.metadata}
            if other_metrics_cols != metadata_cols:
                err_msg = (
                    "The dataframe of metrics passed to MetricsIOManager"
                    " must have the same columns as the metadata"
                    " specifies. The metadata specifies columns"
                    f" {metadata_cols}, but the dataframe has columns"
                    f" {other_metrics_cols}."
                )
                raise ValueError(err_msg)
            # In each tuple, the list of MetricMetadata must all have the same
            # filepath, as the corresponding dataframe is saved to that path
            this_mmd_filepaths = {mmd.metric_parquet_path for mmd in output.metadata}
            if len(this_mmd_filepaths) != 1:
                err_msg = (
                    "The list of MetricMetadata in each tuple passed to"
                    " MetricsIOManager must all have the same"
                    " `metric_parquet_path`."
                )
                raise ValueError(err_msg)
            # Check that it's not already been used
            this_filepath = this_mmd_filepaths.pop()
            if this_filepath in all_filepaths:
                err_msg = (
                    "Each MetricsOutput passed to MetricsIOManager must have a"
                    f" unique `metric_parquet_path`. The path {this_filepath}"
                    " is repeated."
                )
            all_filepaths.append(this_filepath)
        # Convert GEO_ID cols to strings
        for output in obj:
            output.metrics[COL.GEO_ID.value] = output.metrics[COL.GEO_ID.value].astype(
                "string"
            )

        # Aggregate all the MetricMetadatas into a single dataframe, then
        # serialise
        all_metadatas_df = metadata_to_dataframe(
            [m for metrics_output in obj for m in metrics_output.metadata]
        )
        metadata_df_filepath = self.get_full_path_metadata(context)
        self.handle_df(context, all_metadatas_df, metadata_df_filepath)

        # Write dataframes to the parquet files specified in the first element
        # of the tuple
        for metrics_output in obj:
            rel_path = metrics_output.metadata[0].metric_parquet_path
            full_path = self.get_full_path_metrics(rel_path)
            self.handle_df(context, metrics_output.metrics, full_path)

        # Add metadata
        context.add_output_metadata(
            metadata={
                "metric_parquet_paths": all_filepaths,
                "num_metrics": len(all_metadatas_df),
                "metric_human_readable_names": all_metadatas_df[
                    COL.METRIC_HUMAN_READABLE_NAME.value
                ].tolist(),
                "metadata_parquet_path": str(metadata_df_filepath),
                "metadata_preview": MetadataValue.md(
                    all_metadatas_df.head().to_markdown()
                ),
            }
        )


class MetricsPartitionedIOManager(PoppusherIOManager):
    def get_full_path_metrics(
        self,
        parquet_path: str,
    ) -> UPath:
        base_path = self.get_base_path()
        return base_path / UPath(parquet_path)

    def handle_output(
        self,
        context: OutputContext,
        metrics_output: MetricsOutput,
    ) -> None:
        # Check if empty
        if metrics_output.metrics.shape == (0, 0):
            err_msg = (
                "The dataframe of metrics passed to MetricsPartitionedIOManager"
                f" is empty for partition key: {context.partition_key}"
            )
            # TODO: consider whether better approach than raise (error for output)
            # and returning early (no indication directly of empty dataframe
            # aside from log).
            # raise ValueError
            context.log.warning(err_msg)
            return

        # Check GEO_ID col
        metrics_cols = set(metrics_output.metrics.columns)
        if COL.GEO_ID.value not in metrics_cols:
            # Check if it's the index, reset index if so
            if metrics_output.metrics.index.name == COL.GEO_ID.value:
                metrics_output.metrics = metrics_output.metrics.reset_index()
                metrics_cols = set(metrics_output.metrics.columns)
            else:
                err_msg = (
                    "The dataframe of metrics passed to MetricsIOManager"
                    f" must have a {COL.GEO_ID.value} column. It only has columns"
                    f" {metrics_cols}."
                )
                raise ValueError(err_msg)

        # In each tuple, the list of MetricMetadata must all have the same
        # filepath, as the corresponding dataframe is saved to that path
        this_mmd_filepaths = {
            mmd.metric_parquet_path for mmd in metrics_output.metadata
        }
        if len(this_mmd_filepaths) != 1:
            err_msg = (
                "The list of MetricMetadata in each tuple passed to"
                " MetricsIOManager must all have the same"
                " `metric_parquet_path`."
            )
            raise ValueError(err_msg)

        # Check that it's not already been used
        # TODO: this check is not possible for partitioned assets since the
        # data is confined to each partition
        this_filepath = this_mmd_filepaths.pop()

        # Convert GEO_ID cols to strings
        metrics_output.metrics[COL.GEO_ID.value] = metrics_output.metrics[
            COL.GEO_ID.value
        ].astype("string")

        rel_path = metrics_output.metadata[0].metric_parquet_path
        full_path = self.get_full_path_metrics(rel_path)
        self.handle_df(context, metrics_output.metrics, full_path)

        # Add metadata
        context.add_output_metadata(
            metadata={
                "metric_parquet_path": this_filepath,
                "num_metrics": metrics_output.metrics.shape[1] - 1,
            }
        )

    def load_input(self, _context: InputContext) -> pd.DataFrame:
        return super().load_input(_context)


class MetricsMetdataIOManager(PoppusherIOManager):
    def get_full_path_metadata(
        self,
        context: OutputContext,
    ) -> UPath:
        base_path = self.get_base_path()
        asset_prefix = list(context.partition_key.split("/"))[:-1]
        return base_path / UPath("/".join([*asset_prefix, "metric_metadata.parquet"]))

    def handle_output(
        self,
        context: OutputContext,
        obj: list[MetricMetadata],
    ) -> None:
        all_metadatas_df = metadata_to_dataframe(obj)
        metadata_df_filepath = self.get_full_path_metadata(context)
        self.handle_df(context, all_metadatas_df, metadata_df_filepath)

        context.add_output_metadata(
            metadata={
                "num_metrics": all_metadatas_df.shape[0],
                "metric_human_readable_names": all_metadatas_df[
                    COL.METRIC_HUMAN_READABLE_NAME.value
                ].tolist(),
                "metadata_parquet_path": str(metadata_df_filepath),
                "metadata_preview": MetadataValue.md(
                    all_metadatas_df.head().to_markdown()
                ),
            }
        )



from __future__ import annotations

from . import bel, gb_eaw, gb_nir, gb_sct, uk, usa

countries = [
    (mod, mod.__name__.split(".")[-1]) for mod in [bel, gb_nir, uk, usa, gb_eaw, gb_sct]
]

__all__ = ["countries"]



from __future__ import annotations

from datetime import date

import pytest

from poppusher.metadata import (
    COL,
    CountryMetadata,
    DataPublisher,
    GeometryMetadata,
    SourceDataRelease,
)


def test_column_name_uniqueness():
    # Ensure that there are no duplicate column names in the output parquet, so
    # that we don't have issues when merging them in the full catalogue.
    assert len({x.value for x in COL}) == len(COL)


@pytest.mark.xfail()
def test_source_data_release_validation_reference():
    SourceDataRelease(
        name="Test Data Release",
        date_published=date(2021, 1, 1),
        reference_period_start=date(2020, 12, 31),
        reference_period_end=date(2020, 1, 1),
        collection_period_start=date(2020, 1, 1),
        collection_period_end=date(2020, 12, 31),
        expect_next_update=date(2022, 1, 1),
        url="https://example.com",
        data_publisher_id="test_publisher_id",
        description="This is a test data release",
        geometry_metadata_id="test_geom_id",
    )


@pytest.mark.xfail()
def test_source_data_release_validation_collection():
    SourceDataRelease(
        name="Test Data Release",
        date_published=date(2021, 1, 1),
        reference_period_start=date(2020, 1, 1),
        reference_period_end=date(2020, 12, 31),
        collection_period_start=date(2020, 12, 31),
        collection_period_end=date(2020, 1, 1),
        expect_next_update=date(2022, 1, 1),
        url="https://example.com",
        data_publisher_id="test_publisher_id",
        description="This is a test data release",
        geometry_metadata_id="test_geom_id",
    )


def test_source_data_release_hash():
    source_data_release = SourceDataRelease(
        name="Test Data Release",
        date_published=date(2021, 1, 1),
        reference_period_start=date(2020, 1, 1),
        reference_period_end=date(2020, 12, 31),
        collection_period_start=date(2020, 1, 1),
        collection_period_end=date(2020, 12, 31),
        expect_next_update=date(2022, 1, 1),
        url="https://example.com",
        data_publisher_id="test_publisher_id",
        description="This is a test data release",
        geometry_metadata_id="test_geom_id",
    )
    assert (
        source_data_release.id
        == "9ec7e234d73664339e4c1f04bfa485dbb17e204dd72dc3ffbb9cab6870475597"
    )

    source_data_release2 = SourceDataRelease(
        name="Test Data Release2",
        date_published=date(2021, 1, 1),
        reference_period_start=date(2020, 1, 1),
        reference_period_end=date(2020, 12, 31),
        collection_period_start=date(2020, 1, 1),
        collection_period_end=date(2020, 12, 31),
        expect_next_update=date(2022, 1, 1),
        url="https://example.com",
        data_publisher_id="test_publisher_id",
        description="This is a test data release",
        geometry_metadata_id="test_geom_id",
    )
    assert source_data_release.id != source_data_release2.id


def test_data_publisher_hash():
    data_publisher = DataPublisher(
        name="Test Publisher",
        url="https://example.com",
        description="This is a test publisher",
        countries_of_interest=["GBR"],
    )
    assert (
        data_publisher.id
        == "0238fa7ccdc4b5095e62d088a0377bb83e40f62895071f2cc2a75333a98895af"
    )

    data_publisher2 = DataPublisher(
        name="Test Publisher 2",
        url="https://example.com",
        description="This is a test publisher",
        countries_of_interest=["GBR"],
    )
    assert data_publisher.id != data_publisher2.id


def test_geometry_hash():
    country_metadata = CountryMetadata(
        name_short_en="United States",
        name_official="United States of America",
        iso2="US",
        iso3="USA",
        iso3166_2=None,
    )
    geometry_metadata = GeometryMetadata(
        country_metadata=country_metadata,
        validity_period_start=date(2021, 1, 1),
        validity_period_end=date(2021, 1, 1),
        level="tract",
        hxl_tag="tract",
    )
    assert (
        geometry_metadata.id
        == "082cfebd7348ca2d06353ff1d73e6096a60960f9795a26de54faeda777cd7f5d"
    )
    geometry_metadata1 = GeometryMetadata(
        country_metadata=country_metadata,
        validity_period_start=date(2021, 1, 1),
        validity_period_end=date(2021, 1, 1),
        level="tract",
        hxl_tag="tract",
    )
    geometry_metadata2 = GeometryMetadata(
        country_metadata=country_metadata,
        validity_period_start=date(2020, 1, 1),
        validity_period_end=date(2021, 1, 1),
        level="tract",
        hxl_tag="tract",
    )
    geometry_metadata3 = GeometryMetadata(
        country_metadata=country_metadata,
        validity_period_start=date(2021, 1, 1),
        validity_period_end=date(2021, 2, 1),
        level="tract",
        hxl_tag="tract",
    )
    geometry_metadata4 = GeometryMetadata(
        country_metadata=country_metadata,
        validity_period_start=date(2021, 1, 1),
        validity_period_end=date(2021, 1, 1),
        level="block_group",
        hxl_tag="tract",
    )
    geometry_metadata5 = GeometryMetadata(
        country_metadata=country_metadata,
        validity_period_start=date(2021, 1, 1),
        validity_period_end=date(2021, 1, 1),
        level="tract",
        hxl_tag="block_group",
    )
    assert geometry_metadata.id == geometry_metadata1.id
    assert geometry_metadata.id != geometry_metadata2.id
    assert geometry_metadata.id != geometry_metadata3.id
    assert geometry_metadata.id != geometry_metadata4.id
    assert geometry_metadata.id != geometry_metadata5.id
